import uuid

from globus_sdk.scopes import ComputeScopes, FlowsScopes, Scope
from globus_sdk.scopes.collection import (
    DynamicScopeCollection,
    StaticScopeCollection,
    _url_scope,
    _urn_scope,
)


def test_url_scope_string():
    resource_server = str(uuid.UUID(int=0))
    s = _url_scope(resource_server, "data_access")
    assert isinstance(s, Scope)
    assert str(s) == (
        "https://auth.globus.org/scopes/00000000-0000-0000-0000-000000000000"
        "/data_access"
    )


def test_urn_scope_string():
    resource_server = "example.globus.org"
    s = _urn_scope(resource_server, "myscope")
    assert isinstance(s, Scope)
    assert str(s) == "urn:globus:auth:scope:example.globus.org:myscope"


def test_static_scope_collection_iter_contains_expected_values():
    class _MyScopes(StaticScopeCollection):
        resource_server = str(uuid.UUID(int=0))

        foo = _urn_scope(resource_server, "foo")
        bar = _url_scope(resource_server, "bar")

    MyScopes = _MyScopes()

    listified = list(MyScopes)
    as_set = set(listified)
    assert len(listified) == len(as_set)
    assert as_set == {MyScopes.foo, MyScopes.bar}


def test_dynamic_scope_collection_contains_expected_values():
    class MyScopes(DynamicScopeCollection):
        _scope_names = ("foo", "bar")

        @property
        def foo(self):
            return _urn_scope(self.resource_server, "foo")

        @property
        def bar(self):
            return _url_scope(self.resource_server, "bar")

    resource_server = str(uuid.UUID(int=10))
    scope_collection = MyScopes(resource_server)
    assert scope_collection.resource_server == resource_server

    listified = list(scope_collection)
    assert scope_collection.foo in listified
    assert scope_collection.bar in listified


def test_flows_scopes_creation():
    assert FlowsScopes.resource_server == "flows.globus.org"
    assert (
        str(FlowsScopes.run)
        == "https://auth.globus.org/scopes/eec9b274-0c81-4334-bdc2-54e90e689b9a/run"
    )


def test_compute_scopes_creation():
    assert ComputeScopes.resource_server == "funcx_service"
    assert (
        str(ComputeScopes.all)
        == "https://auth.globus.org/scopes/facd7ccc-c5f4-42aa-916b-a0e270e2c2a9/all"
    )



import json

import globus_sdk
from globus_sdk._missing import filter_missing
from globus_sdk.testing import get_last_request, load_response


def test_dummy_timer_creation(client):
    # create a timer with a dummy payload and validate how it gets wrapped by the
    # sending method
    meta = load_response(client.create_timer).metadata

    timer = client.create_timer(timer={"foo": "bar"})
    assert timer["timer"]["job_id"] == meta["timer_id"]

    req = get_last_request()
    sent = json.loads(req.body)
    assert sent == {"timer": {"foo": "bar"}}


def test_transfer_timer_creation(client):
    # create a timer using the payload helpers and confirm that it is serialized as
    # desired
    meta = load_response(client.create_timer).metadata
    body = globus_sdk.TransferData(
        source_endpoint=meta["source_endpoint"],
        destination_endpoint=meta["destination_endpoint"],
    )
    body.add_item("/share/godata/file1.txt", "/~/file1.txt")

    schedule = globus_sdk.RecurringTimerSchedule(
        interval_seconds=60, end={"condition": "iterations", "iterations": 3}
    )

    timer = client.create_timer(
        timer=globus_sdk.TransferTimer(body=body, schedule=schedule)
    )
    assert timer["timer"]["job_id"] == meta["timer_id"]

    req = get_last_request()
    sent = json.loads(req.body)
    assert sent["timer"]["schedule"] == {
        "type": "recurring",
        "interval_seconds": 60,
        "end": {"condition": "iterations", "iterations": 3},
    }
    assert sent["timer"]["body"] == {
        k: [filter_missing(data_val) for data_val in v] if k == "DATA" else v
        for k, v in filter_missing(body).items()
        if k != "skip_activation_check"
    }


def test_flow_timer_creation(client):
    # Setup
    meta = load_response(client.create_timer, case="flow_timer_success").metadata

    # Act
    client.create_timer(
        timer=globus_sdk.FlowTimer(
            flow_id=meta["flow_id"],
            body=meta["callback_body"],
            schedule=meta["schedule"],
        )
    )

    # Verify
    req = get_last_request()
    sent = json.loads(req.body)
    assert sent["timer"]["flow_id"] == meta["flow_id"]
    assert sent["timer"]["body"] == meta["callback_body"]
    assert sent["timer"]["schedule"] == meta["schedule"]



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

metadata = {
    "id": "daa09846-eb92-11e9-b89c-9cb6d0d9fd63",
    "display_name": "example gateway 1",
}

RESPONSES = ResponseSet(
    metadata=metadata,
    default=RegisteredResponse(
        service="gcs",
        method="PATCH",
        path=f"/storage_gateways/{metadata['id']}",
        json={
            "DATA_TYPE": "result#1.0.0",
            "http_response_code": 200,
            "detail": "success",
            "message": "Operation successful",
            "code": "success",
            "data": [
                {
                    "DATA_TYPE": "storage_gateway#1.0.0",
                    "id": metadata["id"],
                    "display_name": metadata["display_name"],
                    "connector_id": "145812c8-decc-41f1-83cf-bb2a85a2a70b",
                    "require_high_assurance": False,
                    "high_assurance": False,
                    "authentication_assurance_timeout": 15840,
                    "authentication_timeout_mins": 15840,
                    "allowed_domains": ["example.edu"],
                    "mapping": "username_without_domain",
                    "restrict_paths": {
                        "DATA_TYPE": "path_restrictions#1.0.0",
                        "read": ["/"],
                    },
                    "policies": {
                        "DATA_TYPE": "posix_storage_gateway#1.0.0",
                        "groups_allow": ["globus"],
                        "groups_deny": ["nonglobus"],
                    },
                    "users_allow": ["user1"],
                    "users_deny": ["user2"],
                }
            ],
        },
    ),
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

INDEX_ID = "60d1160b-f016-40b0-8545-99619865873d"


RESPONSES = ResponseSet(
    metadata={"index_id": INDEX_ID},
    default=RegisteredResponse(
        service="search",
        path=f"/v1/index/{INDEX_ID}/search",
        method="POST",
        json={
            "@datatype": "GSearchResult",
            "@version": "2017-09-01",
            "count": 1,
            "gmeta": [
                {
                    "@datatype": "GMetaResult",
                    "@version": "2019-08-27",
                    "entries": [
                        {
                            "content": {"foo": "bar"},
                            "entry_id": None,
                            "matched_principal_sets": [],
                        }
                    ],
                    "subject": "foo-bar",
                }
            ],
            "has_next_page": True,
            "offset": 0,
            "total": 10,
        },
    ),
)



from globus_sdk.response import IterableResponse


class GetIdentitiesResponse(IterableResponse):
    """
    Response class specific to the Get Identities API

    Provides iteration on the "identities" array in the response.
    """

    default_iter_key = "identities"


class GetIdentityProvidersResponse(IterableResponse):
    """
    Response class specific to the Get Identity Providers API

    Provides iteration on the "identity_providers" array in the response.
    """

    default_iter_key = "identity_providers"



"""
A Globus SDK Sphinx Extension for Autodoc of Class Methods
"""

from __future__ import annotations

import typing as t

from sphinx.application import Sphinx

from .autodoc_hooks import after_autodoc_signature_replace_MISSING_repr
from .directives import (
    AutoMethodList,
    CopyParams,
    EnumerateTestingFixtures,
    ExpandTestingFixture,
    ExternalDocLink,
    ListKnownScopes,
    PaginatedUsage,
)
from .roles import extdoclink_role


def setup(app: Sphinx) -> dict[str, t.Any]:
    app.add_directive("automethodlist", AutoMethodList)
    app.add_directive("listknownscopes", ListKnownScopes)
    app.add_directive("enumeratetestingfixtures", EnumerateTestingFixtures)
    app.add_directive("expandtestfixture", ExpandTestingFixture)
    app.add_directive("extdoclink", ExternalDocLink)
    app.add_directive("paginatedusage", PaginatedUsage)
    app.add_directive("sdk-sphinx-copy-params", CopyParams)

    app.add_role("extdoclink", extdoclink_role)

    app.connect(
        "autodoc-process-signature", after_autodoc_signature_replace_MISSING_repr
    )

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from __future__ import annotations

import random
import string
import uuid
from unittest.mock import Mock

import pytest

import globus_sdk
from globus_sdk import (
    MISSING,
    MissingType,
    OAuthRefreshTokenResponse,
    OAuthTokenResponse,
    Scope,
)
from globus_sdk.scopes.consents import ConsentForest
from globus_sdk.token_storage import (
    MemoryTokenStorage,
    ScopeRequirementsValidator,
    UnchangingIdentityIDValidator,
    ValidatingTokenStorage,
)
from globus_sdk.token_storage.validating_token_storage import (
    IdentityMismatchError,
    MissingIdentityError,
    MissingTokenError,
    UnmetScopeRequirementsError,
)
from tests.common import make_consent_forest


def _make_memstorage_with_scope_validator(consent_client, scope_requirements):
    scope_validator = ScopeRequirementsValidator(scope_requirements, consent_client)
    return ValidatingTokenStorage(MemoryTokenStorage(), validators=(scope_validator,))


def _make_memstorage_with_identity_id_validator():
    identity_validator = UnchangingIdentityIDValidator()
    return ValidatingTokenStorage(
        MemoryTokenStorage(), validators=(identity_validator,)
    )


def test_validating_token_storage_passes_calls_through():
    # test that the major methods are called on the inner storage:
    # - get_token_data
    # - get_token_data_by_resource_server
    # - remove_token_data
    mock_storage = Mock()
    mock_storage.get_token_data_by_resource_server.return_value = {}
    instance = ValidatingTokenStorage(mock_storage)

    mock_storage.get_token_data.assert_not_called()
    instance.get_token_data("foo")
    mock_storage.get_token_data.assert_called_once_with("foo")

    mock_storage.remove_token_data.assert_not_called()
    instance.remove_token_data("foo")
    mock_storage.remove_token_data.assert_called_once_with("foo")

    # small subtlety here:
    # get_token_data_by_resource_server gets called once on init
    # so check that after we invoke it, the call count is 2
    mock_storage.get_token_data_by_resource_server.assert_called_once()
    instance.get_token_data_by_resource_server()
    assert len(mock_storage.get_token_data_by_resource_server.mock_calls) == 2


def test_validating_token_storage_defaults_to_no_validators():
    # basic contract test -- there should be no "out of the box" validators
    instance = ValidatingTokenStorage(MemoryTokenStorage())
    assert len(instance.validators) == 0


def test_validating_token_storage_initially_calls_validators_with_null_identity_id(
    make_token_response,
):
    # assuming there's no initial token data in storage, on the initial store call
    # there will be an identity_id in the token data but not in the "prior" data slot
    mock_validator = Mock()
    instance = ValidatingTokenStorage(
        MemoryTokenStorage(), validators=(mock_validator,)
    )

    identity_id = str(uuid.uuid4())
    instance.store_token_response(make_token_response(identity_id=identity_id))

    mock_validator.before_store.assert_called_once()
    call_args = mock_validator.before_store.call_args
    assert len(call_args.kwargs) == 0
    assert len(call_args.args) == 2
    token_data, context = call_args.args
    assert context.prior_identity_id is None
    assert context.token_data_identity_id == identity_id


def test_validators_are_invoked_even_when_retrieving_empty_data():
    # `get_token-data_by_resource_server` will invoke validators, even if there is no
    # data available
    mock_validator = Mock()
    instance = ValidatingTokenStorage(
        MemoryTokenStorage(), validators=(mock_validator,)
    )

    data = instance.get_token_data_by_resource_server()
    assert data == {}
    mock_validator.after_retrieve.assert_called_once()


def test_validating_token_storage_evaluates_identity_requirements(make_token_response):
    id_a, id_b = str(uuid.uuid4()), str(uuid.uuid4())
    adapter = _make_memstorage_with_identity_id_validator()

    # Seed the adapter with an initial identity.
    assert adapter.identity_id is None
    adapter.store_token_response(make_token_response(identity_id=id_a))
    assert adapter.identity_id == id_a

    # We should be able to store a token with the same identity.
    adapter.store_token_response(make_token_response(identity_id=id_a))

    # We should not be able to store a token with a different identity.
    with pytest.raises(IdentityMismatchError):
        adapter.store_token_response(make_token_response(identity_id=id_b))


def test_validating_token_storage_evaluates_root_scope_requirements(
    make_token_response,
):
    adapter = _make_memstorage_with_scope_validator(
        consent_client, {"rs1": [Scope.parse("scope1")]}
    )
    identity_id = str(uuid.uuid4())
    valid_token_response = make_token_response(
        scopes={"rs1": "scope1"}, identity_id=identity_id
    )
    invalid_token_response = make_token_response(
        scopes={"rs1": "scope2"}, identity_id=identity_id
    )

    adapter.store_token_response(valid_token_response)
    with pytest.raises(UnmetScopeRequirementsError):
        adapter.store_token_response(invalid_token_response)

    assert (
        adapter.get_token_data("rs1").access_token
        == valid_token_response.by_resource_server["rs1"]["access_token"]
    )


def test_storage_with_scope_validator_evaluates_dependent_scope_requirements(
    make_token_response, consent_client
):
    adapter = _make_memstorage_with_scope_validator(
        consent_client, {"rs1": [Scope.parse("scope[subscope]")]}
    )
    token_response = make_token_response(scopes={"rs1": "scope"})
    adapter.store_token_response(token_response)

    consent_client.mocked_forest = make_consent_forest("scope[different_subscope]")
    with pytest.raises(UnmetScopeRequirementsError):
        adapter.get_token_data("rs1")

    consent_client.mocked_forest = make_consent_forest("scope[subscope]")
    adapter.store_token_response(token_response)

    assert (
        adapter.get_token_data("rs1").access_token
        == token_response.by_resource_server["rs1"]["access_token"]
    )


def test_validating_token_storage_fails_non_identifiable_responses(
    make_token_response,
):
    adapter = _make_memstorage_with_identity_id_validator()
    token_response = make_token_response(identity_id=None)

    with pytest.raises(MissingIdentityError):
        adapter.store_token_response(token_response)


def test_validating_token_storage_loads_identity_info_from_storage(
    make_token_response,
):
    # Create an in memory storage adapter
    storage = MemoryTokenStorage()
    adapter = ValidatingTokenStorage(storage)

    # Store an identifiable token response
    identity_id = str(uuid.uuid4())
    token_response = make_token_response(identity_id=identity_id)
    adapter.store_token_response(token_response)

    # Create a net new adapter, pointing at the same storage.
    new_adapter = ValidatingTokenStorage(storage)
    # Verify that the new adapter loads the identity info from storage.
    assert new_adapter.identity_id == identity_id


def test_validating_token_storage_stores_with_saved_identity_id_on_refresh_tokens(
    make_token_response,
):
    # Create an in memory storage adapter with identity_id verification
    adapter = _make_memstorage_with_identity_id_validator()

    # Store an identifiable token response
    identity_id = str(uuid.uuid4())
    token_response = make_token_response(identity_id=identity_id)
    adapter.store_token_response(token_response)

    # now get and store a replacement token response, identified with a different user
    # however, in this case make it a refresh token response
    other_identity_id = str(uuid.uuid4())
    refresh_token_response = make_token_response(
        response_class=OAuthRefreshTokenResponse, identity_id=other_identity_id
    )
    adapter.store_token_response(refresh_token_response)

    # read back the data, and verify that it contains tokens from the refresh, but the
    # original identity_id
    result = adapter.get_token_data("auth.globus.org")
    assert result.access_token == refresh_token_response["access_token"]
    assert result.identity_id == identity_id


def test_validating_token_storage_raises_error_when_no_token_data():
    adapter = ValidatingTokenStorage(MemoryTokenStorage())

    with pytest.raises(MissingTokenError):
        adapter.get_token_data("rs1")


@pytest.fixture
def make_token_response(make_response):
    def _make_token_response(
        scopes: dict[str, str] | None = None,
        identity_id: str | None | MissingType = MISSING,
        response_class: type[OAuthTokenResponse] = OAuthTokenResponse,
    ):
        """
        :param scopes: A dictionary of resource server to scope mappings to fill in
            other tokens.
        :param identity_id: The identity ID to use in the ID token. If None, no ID token
            will be included in the response. If MISSING, the ID token will be generated
            with a random identity ID.
        """
        if scopes is None:
            scopes = {}

        auth_scopes = "openid"
        if "auth.globus.org" in scopes:
            auth_scopes = scopes.pop("auth.globus.org")
            if "openid" not in auth_scopes:
                auth_scopes = f"openid {auth_scopes}"

        data = {
            "access_token": _make_access_token(),
            "expires_in": 172800,
            "other_tokens": [
                {
                    "access_token": _make_access_token(),
                    "expires_in": 172800,
                    "resource_server": resource_server,
                    "scope": scope,
                    "token_type": "Bearer",
                }
                for resource_server, scope in scopes.items()
            ],
            "resource_server": "auth.globus.org",
            "scope": auth_scopes,
            "token_type": "Bearer",
        }

        if identity_id is not None:
            # We'll be mocking out the decode_id_token method, so this doesn't need to
            #   be a real JWT ID token.
            data["id_token"] = _make_id_token()

        response = make_response(response_class=response_class, json_body=data)

        if identity_id is not None:
            decoded_id_token = _decoded_id_token(identity_id)
            response.decode_id_token = lambda: decoded_id_token

        return response

    return _make_token_response


def _decoded_id_token(
    identity_id: str | MissingType = MISSING,
):
    if identity_id is MISSING:
        identity_id = str(uuid.uuid4())
    identity_provider = str(uuid.uuid4())
    aud = str(uuid.uuid4())

    return {
        "at_hash": "RcNb88Asztn-GnRg_0ojS0sSJs1T8YWeYmVkLp7YhdQ",
        "aud": aud,
        "email": None,
        "exp": 1712945792,
        "iat": 1712772992,
        "identity_provider": identity_provider,
        "identity_provider_display_name": "Globus Auth",
        "identity_set": [
            {
                "email": None,
                "identity_provider": identity_provider,
                "identity_provider_display_name": "Globus Auth",
                "last_authentication": None,
                "name": "Pete Sampras",
                "sub": identity_id,
                "username": "kingoftennis@gmail.com",
            }
        ],
        "iss": "https://auth.globus.org",
        "last_authentication": None,
        "name": "Pete Sampras",
        "preferred_username": "kingoftennis@gmail.com",
        "sub": identity_id,
    }


def _make_access_token():
    vocab = string.ascii_letters + string.digits
    return "".join(random.choices(vocab, k=91))


def _make_id_token():
    return "".join(random.choices(string.ascii_letters, k=1000))


@pytest.fixture
def consent_client() -> MockedConsentClient:
    client = Mock(spec=MockedConsentClient)
    client.mocked_forest = None
    get_consents_response_mock = Mock(spec=globus_sdk.GetConsentsResponse)

    def retrieve_mocked_forest():
        if client.mocked_forest is None:
            raise ValueError("No mocked_forest has been set on the client")
        return client.mocked_forest

    get_consents_response_mock.to_forest.side_effect = retrieve_mocked_forest
    client.get_consents.return_value = get_consents_response_mock
    return client


class MockedConsentClient(globus_sdk.AuthClient):
    mocked_forest: ConsentForest | None



import types
import typing as t

from docutils.parsers.rst import directives

from ..utils import classname2methods
from .add_content_directive import AddContentDirective


class AutoMethodList(AddContentDirective):
    has_content = False
    required_arguments = 1
    optional_arguments = 0
    option_spec = {"include_methods": directives.unchanged}

    def gen_rst(self) -> t.Iterator[str]:
        classname = self.arguments[0]

        include_methods = []
        if "include_methods" in self.options:
            include_methods = self.options["include_methods"].strip().split(",")

        yield ""
        yield "**Methods**"
        yield ""
        for methodname, method in classname2methods(classname, include_methods):
            if not is_paginated_method(method):
                yield f"* :py:meth:`~{classname}.{methodname}`"
            else:
                yield (
                    f"* :py:meth:`~{classname}.{methodname}`, "
                    f"``paginated.{methodname}()``"
                )

        yield ""


def is_paginated_method(func: types.FunctionType) -> bool:
    return getattr(func, "_has_paginator", False)



from __future__ import annotations

import collections.abc
import json
import logging
import typing as t
from functools import cached_property

from globus_sdk._internal import guards

log = logging.getLogger(__name__)

if t.TYPE_CHECKING:
    from requests import Response

    import globus_sdk

__all__ = (
    "GlobusHTTPResponse",
    "IterableResponse",
    "ArrayResponse",
)


class GlobusHTTPResponse:
    """
    Response object that wraps an HTTP response from the underlying HTTP
    library. If the response is JSON, the parsed data will be available in
    ``data``, otherwise ``data`` will be ``None`` and ``text`` should
    be used instead.

    The most common response data is a JSON dictionary. To make
    handling this type of response as seamless as possible, the
    ``GlobusHTTPResponse`` object implements the immutable mapping protocol for
    dict-style access. This is just an alias for access to the underlying data.

    If the response data is not a dictionary or list, item access will raise
    ``TypeError``.

    >>> print("Response ID": r["id"]) # alias for r.data["id"]

    :ivar client: The client instance which made the request
    """

    def __init__(
        self,
        response: Response | GlobusHTTPResponse,
        client: globus_sdk.BaseClient | None = None,
    ) -> None:
        # init on a GlobusHTTPResponse: we are wrapping this data
        # the _response is None
        if isinstance(response, GlobusHTTPResponse):
            if client is not None:
                raise ValueError("Redundant client with wrapped response")
            self._wrapped: GlobusHTTPResponse | None = response
            self._response: Response | None = None
            self.client: globus_sdk.BaseClient = self._wrapped.client

        # init on a Response object, this is the "normal" case
        # _wrapped is None
        else:
            if client is None:
                raise ValueError("Missing client with normal response")
            self._wrapped = None
            self._response = response
            self.client = client

    @cached_property
    def _parsed_json(self) -> t.Any:
        # JSON decoding may raise a ValueError due to an invalid JSON
        # document. In the case of trying to fetch the "data" on an HTTP
        # response, this means we didn't get a JSON response.
        # store this as None, as in "no data"
        #
        # if the caller *really* wants the raw body of the response, they can
        # always use `text`
        if self._wrapped is not None:
            return self._wrapped._parsed_json

        if self._response is not None:
            try:
                return self._response.json()
            except ValueError:
                log.warning("response data did not parse as JSON, data=None")
                return None

        raise NotImplementedError(
            "Cannot read JSON data from a response which did not "
            "wrap another response or contain a transport-layer response."
        )

    @property
    def _raw_response(self) -> Response:
        # this is an internal property which traverses any series of wrapped responses
        # until reaching a requests response object
        if self._response is not None:
            return self._response
        elif self._wrapped is not None:
            return self._wrapped._raw_response
        else:  # unreachable  # pragma: no cover
            raise ValueError("could not find an inner response object")

    @property
    def http_status(self) -> int:
        """The HTTP response status, as an integer."""
        return self._raw_response.status_code

    @property
    def http_reason(self) -> str:
        """
        The HTTP reason string from the response.

        This is the part of the status line after the status code, and typically is a
        string description of the status. If the status line is
        ``HTTP/1.1 200 OK``, then this is the string ``"OK"``.
        """
        return self._raw_response.reason

    @property
    def headers(self) -> t.Mapping[str, str]:
        """
        The HTTP response headers as a case-insensitive mapping.

        For example, ``headers["Content-Length"]`` and ``headers["content-length"]`` are
        treated as equivalent.
        """
        return self._raw_response.headers

    @property
    def content_type(self) -> str | None:
        return self.headers.get("Content-Type")

    @property
    def text(self) -> str:
        """The raw response data as a string."""
        return self._raw_response.text

    @property
    def binary_content(self) -> bytes:
        """
        The raw response data in bytes.
        """
        return self._raw_response.content

    @property
    def data(self) -> t.Any:
        return self._parsed_json

    def get(self, key: str, default: t.Any = None) -> t.Any:
        """
        ``get`` is just an alias for ``data.get(key, default)``, but with the added
        checks that if ``data`` is ``None`` or a list, it returns the default.

        :param key: The string key to lookup in the response if it is a dict
        :param default: The default value to be used if the data is null or a list
        """
        if guards.is_optional(self.data, list):
            return default
        # NB: `default` is provided as a positional because the native dict type
        # doesn't recognize a keyword argument `default`
        return self.data.get(key, default)

    def __str__(self) -> str:
        """The default __str__ for a response assumes that the data is valid
        JSON-dump-able."""
        if self.data is not None:
            return json.dumps(self.data, indent=2, separators=(",", ": "))
        return self.text

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.text})"

    def __getitem__(self, key: str | int | slice) -> t.Any:
        # force evaluation of the data property outside of the upcoming
        # try-catch so that we don't accidentally catch TypeErrors thrown
        # during the getter function itself
        data = self.data
        try:
            return data[key]
        except TypeError as err:
            log.error(
                f"Can't index into responses with underlying data of type {type(data)}"
            )
            # re-raise with an altered message and error type -- the issue is that
            # whatever data is in the response doesn't support indexing (e.g. a response
            # that is just an integer, parsed as json)
            #
            # "type" is ambiguous, but we don't know if it's the fault of the
            # class at large, or just a particular call's `data` property
            raise ValueError(
                "This type of response data does not support indexing."
            ) from err

    def __contains__(self, item: t.Any) -> bool:
        """
        ``x in response`` is an alias for ``x in response.data``
        """
        if self.data is None:
            return False
        return item in self.data

    def __bool__(self) -> bool:
        """
        ``bool(response)`` is an alias for ``bool(response.data)``
        """
        return bool(self.data)


class IterableResponse(GlobusHTTPResponse):
    """This response class adds an __iter__ method on an 'iter_key' variable.
    The assumption is that iter produces dicts or dict-like mappings."""

    default_iter_key: t.ClassVar[str]
    iter_key: str

    def __init__(
        self,
        response: Response | GlobusHTTPResponse,
        client: globus_sdk.BaseClient | None = None,
        *,
        iter_key: str | None = None,
    ) -> None:
        if not hasattr(self, "default_iter_key"):
            raise TypeError(
                "Cannot instantiate an iterable response from a class "
                "which does not define a default iteration key."
            )
        iter_key = iter_key if iter_key is not None else self.default_iter_key
        self.iter_key = iter_key
        super().__init__(response, client)

    def __iter__(self) -> t.Iterator[t.Mapping[t.Any, t.Any]]:
        if not isinstance(self.data, dict):
            raise TypeError(
                "Cannot iterate on IterableResponse data when "
                f"type is '{type(self.data).__name__}'"
            )
        return iter(self.data[self.iter_key])


class ArrayResponse(GlobusHTTPResponse):
    """This response class adds an ``__iter__`` method which assumes that the top-level
    data of the response is a JSON array."""

    def __iter__(self) -> t.Iterator[t.Any]:
        if not isinstance(self.data, list):
            raise TypeError(
                "Cannot iterate on ArrayResponse data when "
                f"type is '{type(self.data).__name__}'"
            )
        return iter(self.data)

    def __len__(self) -> int:
        """
        ``len(response)`` is an alias for ``len(response.data)``
        """
        if not isinstance(self.data, collections.abc.Sequence):
            raise TypeError(
                "Cannot take len() on ArrayResponse data when "
                f"type is '{type(self.data).__name__}'"
            )
        return len(self.data)


class IterableJSONAPIResponse(IterableResponse):
    """This response class is for iterating on JSON:API responses that have an
    array of data in their top level `data` key."""

    default_iter_key = "data"



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from .._common import FUNCTION_CODE, FUNCTION_ID, FUNCTION_NAME

FUNCTION_DOC = {
    "function_uuid": FUNCTION_ID,
    "function_name": FUNCTION_NAME,
    "function_code": FUNCTION_CODE,
    "description": "I just wanted to say hello.",
    "metadata": {"python_version": "3.12.6", "sdk_version": "2.28.1"},
}

RESPONSES = ResponseSet(
    metadata={
        "function_id": FUNCTION_ID,
        "function_name": FUNCTION_NAME,
        "function_code": FUNCTION_CODE,
    },
    default=RegisteredResponse(
        service="compute",
        path=f"/v2/functions/{FUNCTION_ID}",
        method="GET",
        json=FUNCTION_DOC,
    ),
)



from globus_sdk import response


class IterableFlowsResponse(response.IterableResponse):
    """
    An iterable response containing a "flows" array of flow definitions.

    This response type is returned by :meth:`FlowsClient.list_flows` and provides
    iteration over individual flow objects from a single page of results.

    When iterated over, yields individual flow dictionaries, where each flow
    typically contains:

    - ``id``: UUID of the flow
    - ``title``: Display title of the flow
    - ``definition``: The flow's JSON definition
    - ``input_schema``: JSON Schema for flow inputs
    - ``globus_auth_scope``: Auth scope string for the flow
    - ``created_at``: Timestamp of flow creation
    - ``updated_at``: Timestamp of last update
    """

    default_iter_key = "flows"


class IterableRunsResponse(response.IterableResponse):
    """
    An iterable response containing a "runs" array of flow run records.

    This response type is returned by :meth:`FlowsClient.list_runs` and provides
    iteration over individual run objects from a single page of results.

    When iterated over, yields individual run dictionaries, where each run
    typically contains:

    - ``run_id``: UUID of the run
    - ``flow_id``: UUID of the flow this run belongs to
    - ``flow_title``: Title of the flow
    - ``status``: Current status of the run (e.g., "ACTIVE", "INACTIVE",
      "SUCCEEDED", "FAILED", "ENDED")
    - ``start_time``: Timestamp when the run was started
    - ``completion_time``: Timestamp when the run completed (if applicable)
    - ``run_owner``: Identity URN of the user who started the run
    """

    default_iter_key = "runs"


class IterableRunLogsResponse(response.IterableResponse):
    """
    An iterable response containing an "entries" array of log entries.

    This response type is returned by :meth:`FlowsClient.get_run_logs` and provides
    iteration over individual log entries from a single page of results.

    When iterated over, yields individual log entry dictionaries, where each entry
    typically contains:

    - ``time``: Timestamp of the log entry
    - ``code``: Event code (e.g., "FlowStarted", "ActionStarted", "ActionCompleted")
    - ``description``: Human-readable description of the event
    - ``details``: Additional nested information about the event
    """

    default_iter_key = "entries"


class IterableRegisteredAPIsResponse(response.IterableResponse):
    """
    An iterable response containing a "registered_apis" array of registered API records.

    This response type is returned by :meth:`FlowsClient.list_registered_apis` and
    provides iteration over individual registered API objects from a single page of
    results.

    When iterated over, yields individual registered API dictionaries, where each
    registered API typically contains:

    - ``id``: UUID of the registered API
    - ``name``: Display name of the registered API
    - ``description``: Description of the registered API
    - ``created_timestamp``: Timestamp of registered API creation
    - ``updated_timestamp``: Timestamp of last update
    """

    default_iter_key = "registered_apis"



from __future__ import annotations

import json
import pathlib
import sqlite3
import textwrap
import typing as t

from globus_sdk import __version__, exc

from .base import FileTokenStorage
from .token_data import TokenStorageData


class SQLiteTokenStorage(FileTokenStorage):
    """
    A token storage which stores token data on disk in a SQLite database.

    See :class:`TokenStorage` for common interface details.

    :param filepath: The path on disk to a SQLite database file.
    :param connect_params: A dictionary of parameters to pass to ``sqlite3.connect()``.
    :param namespace: A unique string for partitioning token data (Default: "DEFAULT").

    :raises GlobusSDKUsageError: If the filepath is ":memory:". This usage-mode is not
        supported in this class; use :class:`MemoryTokenStorage` instead if in-memory
        token storage is desired.
    """

    file_format = "db"

    def __init__(
        self,
        filepath: pathlib.Path | str,
        *,
        connect_params: dict[str, t.Any] | None = None,
        namespace: str = "DEFAULT",
    ) -> None:
        if filepath == ":memory:":
            raise exc.GlobusSDKUsageError(
                "SQLiteTokenStorage cannot be used with a ':memory:' database. "
                "If you want to store tokens in memory, use MemoryTokenStorage instead."
            )

        super().__init__(filepath, namespace=namespace)
        self._connection = self._init_and_connect(connect_params)

    def _init_and_connect(
        self,
        connect_params: dict[str, t.Any] | None,
    ) -> sqlite3.Connection:
        connect_params = connect_params or {}
        if not self.file_exists():
            with self.user_only_umask():
                conn: sqlite3.Connection = sqlite3.connect(
                    self.filepath, **connect_params
                )
            conn.executescript(textwrap.dedent("""
                    CREATE TABLE token_storage (
                        namespace VARCHAR NOT NULL,
                        resource_server VARCHAR NOT NULL,
                        token_data_json VARCHAR NOT NULL,
                        PRIMARY KEY (namespace, resource_server)
                    );
                    CREATE TABLE sdk_storage_adapter_internal (
                        attribute VARCHAR NOT NULL,
                        value VARCHAR NOT NULL,
                        PRIMARY KEY (attribute)
                    );
                    """))
            # mark the version which was used to create the DB
            # also mark the "database schema version" in case we ever need to handle
            # graceful upgrades
            conn.executemany(
                "INSERT INTO sdk_storage_adapter_internal(attribute, value) "
                "VALUES (?, ?)",
                [
                    ("globus-sdk.version", __version__),
                    # schema_version=1 indicates a schema built with the original
                    # SQLiteAdapter
                    # schema_version=2 indicates one built with SQLiteTokenStorage
                    #
                    # a schema_version of 1 therefore indicates that there should be
                    # a 'config_storage' table present
                    ("globus-sdk.database_schema_version", "2"),
                ],
            )
            conn.commit()
        else:
            conn = sqlite3.connect(self.filepath, **connect_params)
        return conn

    def close(self) -> None:
        """
        Close the underlying database connection.
        """
        self._connection.close()

    def store_token_data_by_resource_server(
        self, token_data_by_resource_server: t.Mapping[str, TokenStorageData]
    ) -> None:
        """
        Store token data for one or more resource servers in the current namespace.

        Token data is JSON-serialized before being inserted into the database.

        :param token_data_by_resource_server: A mapping of resource servers to token
            data.
        """
        pairs = []
        for resource_server, token_data in token_data_by_resource_server.items():
            pairs.append((resource_server, token_data.to_dict()))

        self._connection.executemany(
            "REPLACE INTO token_storage(namespace, resource_server, token_data_json) "
            "VALUES(?, ?, ?)",
            [
                (self.namespace, rs_name, json.dumps(token_data_dict))
                for (rs_name, token_data_dict) in pairs
            ],
        )
        self._connection.commit()

    def get_token_data_by_resource_server(self) -> dict[str, TokenStorageData]:
        """
        Lookup all token data under the current namespace from the database.

        :returns: A dict of ``TokenStorageData`` objects indexed by their
            resource server.
        """
        ret: dict[str, TokenStorageData] = {}
        for row in self._connection.execute(
            "SELECT resource_server, token_data_json "
            "FROM token_storage WHERE namespace=?",
            (self.namespace,),
        ):
            resource_server, token_data_json = row
            token_data_dict = json.loads(token_data_json)
            ret[resource_server] = TokenStorageData.from_dict(token_data_dict)
        return ret

    def remove_token_data(self, resource_server: str) -> bool:
        """
        Given a resource server to target, delete token data for that resource server
        from the database (limited to the current namespace).
        You can use this as part of a logout command implementation, loading token data
        as a dict, and then deleting the data for each resource server.

        :param resource_server: The name of the resource server to remove from the DB
        :returns: True if token data was deleted, False if none was found to delete.
        """
        rowcount = self._connection.execute(
            "DELETE FROM token_storage WHERE namespace=? AND resource_server=?",
            (self.namespace, resource_server),
        ).rowcount
        self._connection.commit()
        return rowcount != 0

    def iter_namespaces(self) -> t.Iterator[str]:
        """Iterate over all distinct namespaces in the SQLite database."""
        seen: set[str] = set()
        for row in self._connection.execute(
            "SELECT DISTINCT namespace FROM token_storage;"
        ):
            namespace = row[0]
            seen.add(namespace)
            yield namespace



from globus_sdk.response import IterableResponse


class GetClientsResponse(IterableResponse):
    """
    Response class specific to the Get Clients API

    Provides iteration on the "clients" array in the response.
    """

    default_iter_key = "clients"



import json
import uuid

import pytest

from globus_sdk._missing import MISSING, filter_missing
from globus_sdk.testing import get_last_request, load_response


@pytest.mark.parametrize(
    "admin_id_style", ("missing", "string", "list", "set", "uuid", "uuid_list")
)
def test_update_project_admin_id_styles(service_client, admin_id_style):
    meta = load_response(service_client.update_project).metadata

    if admin_id_style == "missing":
        admin_ids = MISSING
    elif admin_id_style == "string":
        admin_ids = meta["admin_id"]
    elif admin_id_style == "list":
        admin_ids = [meta["admin_id"]]
    elif admin_id_style == "set":
        admin_ids = {meta["admin_id"]}
    elif admin_id_style == "uuid":
        admin_ids = uuid.UUID(meta["admin_id"])
    elif admin_id_style == "uuid_list":
        admin_ids = [uuid.UUID(meta["admin_id"])]
    else:
        raise NotImplementedError(f"unknown admin_id_style {admin_id_style}")

    project_id = meta["id"]
    res = service_client.update_project(
        project_id, display_name="My Project", admin_ids=admin_ids
    )

    assert res["project"]["id"] == meta["id"]

    last_req = get_last_request()
    data = json.loads(last_req.body)
    assert list(data) == ["project"], data  # 'project' is the only key
    if admin_id_style == "missing":
        assert filter_missing(data["project"]) == {"display_name": "My Project"}
    else:
        assert data["project"] == {
            "display_name": "My Project",
            "admin_ids": [meta["admin_id"]],
        }


@pytest.mark.parametrize(
    "group_id_style", ("string", "list", "set", "uuid", "uuid_list")
)
def test_update_project_group_id_styles(service_client, group_id_style):
    meta = load_response(service_client.update_project, case="admin_group").metadata

    if group_id_style == "string":
        group_ids = meta["admin_group_id"]
    elif group_id_style == "list":
        group_ids = [meta["admin_group_id"]]
    elif group_id_style == "set":
        group_ids = {meta["admin_group_id"]}
    elif group_id_style == "uuid":
        group_ids = uuid.UUID(meta["admin_group_id"])
    elif group_id_style == "uuid_list":
        group_ids = [uuid.UUID(meta["admin_group_id"])]
    else:
        raise NotImplementedError(f"unknown group_id_style {group_id_style}")

    project_id = meta["id"]
    res = service_client.update_project(
        project_id, contact_email="support@globus.org", admin_group_ids=group_ids
    )

    assert res["project"]["id"] == meta["id"]

    last_req = get_last_request()
    data = json.loads(last_req.body)
    assert list(data) == ["project"], data  # 'project' is the only key
    assert data["project"] == {
        "contact_email": "support@globus.org",
        "admin_group_ids": [meta["admin_group_id"]],
    }



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import ENDPOINT_ID

ENDPOINT_DOC = {
    "DATA_TYPE": "endpoint",
    "id": ENDPOINT_ID,
    "display_name": "myserver",
    "organization": "My Org",
    "username": "auser",
    "description": "Example gridftp endpoint.",
    "entity_type": "GCSv4_host",
    "public": False,
    "french_english_bilingual": False,
    "is_globus_connect": False,
    "globus_connect_setup_key": None,
    "gcp_connected": None,
    "gcp_paused": None,
    "s3_url": None,
    "s3_owner_activated": False,
    "host_endpoint_id": None,
    "host_path": None,
    "disable_verify": False,
    "disable_anonymous_writes": False,
    "force_verify": False,
    "force_encryption": False,
    "mfa_required": False,
    "myproxy_server": None,
    "myproxy_dn": None,
    "non_functional": False,
    "non_functional_endpoint_display_name": None,
    "non_functional_endpoint_id": None,
    "oauth_server": None,
    "default_directory": None,
    "activated": False,
    "expires_in": 0,
    "expire_time": "2000-01-02 03:45:06+00:00",
    "shareable": True,
    "acl_available": False,
    "acl_editable": False,
    "in_use": False,
    "DATA": [
        {
            "DATA_TYPE": "server",
            "hostname": "gridftp.example.org",
            "uri": "gsiftp://gridftp.example.org:2811",
            "port": 2811,
            "scheme": "gsiftp",
            "id": 985,
            "subject": "/O=Grid/OU=Example/CN=host/gridftp.example.org",
        }
    ],
}


RESPONSES = ResponseSet(
    metadata={"endpoint_id": ENDPOINT_ID},
    default=RegisteredResponse(
        service="transfer",
        path=f"/v0.10/endpoint/{ENDPOINT_ID}",
        json=ENDPOINT_DOC,
    ),
)



from __future__ import annotations

import enum
import logging
import sys
import textwrap
import typing as t

from globus_sdk._internal import guards

from .base import GlobusError
from .err_info import ErrorInfoContainer

if t.TYPE_CHECKING:
    import requests

log = logging.getLogger(__name__)

_CACHE_SENTINEL = object()


class _ErrorFormat(enum.Enum):
    undefined = enum.auto()
    jsonapi = enum.auto()
    type_zero = enum.auto()


class GlobusAPIError(GlobusError):
    """
    Wraps errors returned by a REST API.

    :ivar int http_status: HTTP status code
    :ivar str code: Error code from the API or ``None`` for unclassified errors
    :ivar str request_id: The 'request_id' included in the error data, if any.
    :ivar list[str] messages: A list of error messages, extracted from the response
        data. If the data cannot be parsed or does not contain any clear message fields,
        this list may be empty.
    :ivar list[GlobusSubError] errors: A list of sub-error documents, as would be
        presented by JSON:API APIs and similar interfaces.
    """

    MESSAGE_FIELDS = ["message", "detail", "title"]
    RECOGNIZED_AUTHZ_SCHEMES = ["bearer", "basic", "globus-goauthtoken"]

    def __init__(self, r: requests.Response, *args: t.Any, **kwargs: t.Any) -> None:
        self._cached_raw_json: t.Any = _CACHE_SENTINEL

        self.http_status = r.status_code
        # defaults, may be rewritten during parsing
        self.code: str | None = None
        self.request_id: str | None = None
        self.messages: list[str] = []
        self.errors: list[ErrorSubdocument] = []

        self._info: ErrorInfoContainer | None = None
        self._underlying_response = r
        self._parse_response()

        if sys.version_info >= (3, 11):
            self.add_note(  # pylint: disable=no-member
                (
                    "This exception was caused by an API error. "
                    "The response body is as follows:\n\n"
                )
                + textwrap.indent(self.text, "    ")
            )
        super().__init__(*self._get_args())

    @property
    def message(self) -> str | None:
        """
        An error message from the API.

        If there are multiple messages available, this will contain all messages joined
        with semicolons. If there is no message available, this will be ``None``.
        """
        if self.messages:
            return "; ".join(self.messages)
        return None

    @property
    def http_reason(self) -> str:
        """
        The HTTP reason string from the response.

        This is the part of the status line after the status code, and typically is a
        string description of the status. If the status line is
        ``HTTP/1.1 404 Not Found``, then this is the string ``"Not Found"``.
        """
        return self._underlying_response.reason

    @property
    def headers(self) -> t.Mapping[str, str]:
        """
        The HTTP response headers as a case-insensitive mapping.

        For example, ``headers["Content-Length"]`` and ``headers["content-length"]`` are
        treated as equivalent.
        """
        return self._underlying_response.headers

    @property
    def content_type(self) -> str | None:
        return self.headers.get("Content-Type")

    def _get_mimetype(self, content_type: str) -> str:
        return content_type.split(";")[0].strip()

    def _jsonapi_mimetype(self) -> bool:
        if self.content_type is None:
            return False
        return self._get_mimetype(self.content_type) == "application/vnd.api+json"

    def _json_mimetype(self) -> bool:
        if self.content_type is None:
            return False
        mimetype = self._get_mimetype(self.content_type)
        if mimetype == "application/json":
            return True
        if mimetype.startswith("application/") and mimetype.endswith("+json"):
            return True
        return False

    @property
    def raw_json(self) -> dict[str, t.Any] | None:
        """
        Get the verbatim error message received from a Globus API, interpreted
        as JSON data

        If the body cannot be loaded as JSON, this is None
        """
        if self._cached_raw_json == _CACHE_SENTINEL:
            self._cached_raw_json = None
            if self._json_mimetype():
                try:
                    # technically, this could be a non-dict JSON type, like a list or
                    # string but in those cases the user can just cast -- the "normal"
                    # case is a dict
                    self._cached_raw_json = self._underlying_response.json()
                except ValueError:
                    log.error(
                        "Error body could not be JSON decoded! "
                        "This means the Content-Type is wrong, or the "
                        "body is malformed!"
                    )
        return t.cast("dict[str, t.Any]", self._cached_raw_json)

    @property
    def _dict_data(self) -> dict[str, t.Any]:
        """
        A "type asserting" wrapper over raw_json which errors if the type is not dict.
        """
        if not isinstance(self.raw_json, dict):
            raise ValueError("cannot use _dict_data when data is non-dict type")
        return self.raw_json

    @property
    def text(self) -> str:
        """
        Get the verbatim error message received from a Globus API as a *string*
        """
        return self._underlying_response.text

    @property
    def binary_content(self) -> bytes:
        """
        The error message received from a Globus API in bytes.
        """
        return self._underlying_response.content

    @property
    def info(self) -> ErrorInfoContainer:
        """
        An ``ErrorInfoContainer`` with parsed error data. The ``info`` of an error is
        guaranteed to be present, but all of its contents may be falsey if the error
        could not be parsed.
        """
        if self._info is None:
            json_data = self.raw_json if isinstance(self.raw_json, dict) else None
            self._info = ErrorInfoContainer(json_data)
        return self._info

    def _get_request_authorization_scheme(self) -> str | None:
        try:
            authz_h = self._underlying_response.request.headers["Authorization"]
            authz_scheme = authz_h.split()[0]
            if authz_scheme.lower() in self.RECOGNIZED_AUTHZ_SCHEMES:
                return authz_scheme
        except (IndexError, KeyError):
            pass
        return None

    def _get_args(self) -> list[t.Any]:
        """
        Get arguments to pass to the Exception base class. These args are
        displayed in stack traces.
        """
        args = [
            self._underlying_response.request.method,
            self._underlying_response.url,
            self._get_request_authorization_scheme(),
            self.http_status,
            self.code,
            # if the message is "", try using response reason
            # for details on these, and some examples, see
            #   https://datatracker.ietf.org/doc/html/rfc7231#section-6.1
            self.message or self._underlying_response.reason,
        ]
        if self.request_id:
            args.append(self.request_id)
        return args

    def _parse_response(self) -> bool:
        """
        This is an intermediate step between 'raw_json' (loading bare JSON data)
        and the "real" parsing methods.

        In order to better support subclassing with short-circuiting behaviors if
        parsing goes awry, all of the parsing methods return True if parsing should
        continue or False if it was aborted.

        _parse_response() pulls the JSON body and does the following:

        - on non-dict JSON data, log and abort early. Don't error since this
          is already part of exception construction, just stop parsing.

        - Attempt to detect the error format in use, then dispatch to the relevant
          subparser: JSON:API, Type Zero, or Undefined

        - if subparsing succeeded, call the `_post_parse_hook` hook to allow
          subclasses to trivially add more computed attributes. This could also be
          done by altering `__init__` but it's nicer to have a dedicated hook because
          it is guaranteed to only be called if the rest of parsing succeeded
        """
        if self.raw_json is None:
            log.debug("Error body was not parsed as JSON")
            return False
        if not isinstance(self.raw_json, dict):
            log.warning(  # type: ignore[unreachable]
                "Error body could not be parsed as JSON because it was not a dict"
            )
            return False

        error_format = self._detect_error_format()
        subparse_result: bool
        if error_format == _ErrorFormat.jsonapi:
            subparse_result = self._parse_jsonapi_error_format()
        elif error_format == _ErrorFormat.type_zero:
            subparse_result = self._parse_type_zero_error_format()
        else:
            subparse_result = self._parse_undefined_error_format()

        if not subparse_result:
            return False
        return self._post_parse_hook()

    def _post_parse_hook(self) -> bool:
        """
        An internal callback for extra customizations after fully successful parsing.

        By default, does nothing.
        """
        return True

    def _detect_error_format(self) -> _ErrorFormat:
        # if the JSON:API mimetype was used, inspect the data to make sure it is
        # well-formed
        if self._jsonapi_mimetype():
            errors = self._dict_data.get("errors")
            if not guards.is_list_of(errors, dict):
                return _ErrorFormat.undefined
            elif len(errors) < 1:
                return _ErrorFormat.undefined
            elif not all(isinstance(error_doc, dict) for error_doc in errors):
                return _ErrorFormat.undefined

            # only use 'jsonapi' if everything checked out
            return _ErrorFormat.jsonapi

        # now evaluate attributes for Type Zero errors under the same paradigm
        # check each attribute and only return 'type_zero' if nothing failed

        if not isinstance(self._dict_data.get("code"), str):
            return _ErrorFormat.undefined
        elif not isinstance(self._dict_data.get("message"), str):
            return _ErrorFormat.undefined
        # request_id is not required, but must be a string if present
        elif "request_id" in self._dict_data and not isinstance(
            self._dict_data["request_id"], str
        ):
            return _ErrorFormat.undefined

        return _ErrorFormat.type_zero

    def _parse_jsonapi_error_format(self) -> bool:
        """
        Parsing a JSON:API Error

        This is only called after the field type for 'errors' has been checked.
        However, the nested/underlying fields will not have been checked yet.
        """
        self.errors = [ErrorSubdocument(e) for e in self._dict_data["errors"]]
        self.code = self._extract_code_from_error_array(self.errors)
        self.messages = self._extract_messages_from_error_array(self.errors)
        return True

    def _parse_type_zero_error_format(self) -> bool:
        """
        Parsing a Type Zero Error

        This is only called after Type Zero has been detected. Therefore, we already
        have assurances about the values in 'code' and 'message'.
        Note that 'request_id' could be absent but *must* be a string if present.
        """
        self.code = self._dict_data["code"]
        self.messages = [self._dict_data["message"]]
        self.request_id = self._dict_data.get("request_id")
        if guards.is_list_of(self._dict_data.get("errors"), dict):
            raw_errors = self._dict_data["errors"]
        else:
            raw_errors = [self._dict_data]
        self.errors = [
            ErrorSubdocument(e, message_fields=self.MESSAGE_FIELDS) for e in raw_errors
        ]
        return True

    def _parse_undefined_error_format(self) -> bool:
        """
        Undefined Parsing: best effort support for unknown data shapes

        This is also a great place for custom parsing to hook in for different APIs
        if we know that there's an unusual format in use
        """

        # attempt to pull out errors if possible and valid
        if guards.is_list_of(self._dict_data.get("errors"), dict):
            raw_errors = self._dict_data["errors"]
        # if no 'errors' were found, or 'errors' is invalid, then
        # 'errors' should be set to contain the root document
        else:
            raw_errors = [self._dict_data]
        self.errors = [
            ErrorSubdocument(e, message_fields=self.MESSAGE_FIELDS) for e in raw_errors
        ]

        # use 'code' if present and correct type
        if isinstance(self._dict_data.get("code"), str):
            self.code = t.cast(str, self._dict_data["code"])
        # otherwise, pull 'code' from the sub-errors array
        elif self.errors:
            # in undefined parse cases, the code will be left as `"Error"` for
            # a higher degree of backwards compatibility
            maybe_code = self._extract_code_from_error_array(self.errors)
            if maybe_code is not None:
                self.code = maybe_code

        # either there is an array of subdocument errors or there is not,
        # in which case we load only from the root doc
        self.messages = self._extract_messages_from_error_array(
            self.errors
            or [ErrorSubdocument(self._dict_data, message_fields=self.MESSAGE_FIELDS)]
        )

        return True

    def _extract_messages_from_error_array(
        self, errors: list[ErrorSubdocument]
    ) -> list[str]:
        """
        Extract 'messages' from an array of errors (JSON:API or otherwise)

        Each subdocument *may* define its messages, so this is the aggregate of messages
        from those documents which had messages.

        Note that subdocuments may be instructed about their `message_fields` by the
        error class and parsing path which they take. Therefore, this may be extracting
        `"message"`, `"detail"`, or `"title"` in the base implementation and other
        fields if a subclass customizes this further.
        """
        ret: list[str] = []
        for doc in errors:
            if doc.message is not None:
                ret.append(doc.message)
        return ret

    def _extract_code_from_error_array(
        self, errors: list[ErrorSubdocument]
    ) -> str | None:
        """
        Extract a 'code' field from an array of errors (JSON:API or otherwise)

        This is done by checking if each error document has the same 'code'
        """
        codes: set[str] = set()
        for error in errors:
            if error.code is not None:
                codes.add(error.code)
        if len(codes) == 1:
            return codes.pop()
        return None


class ErrorSubdocument:
    """
    Error subdocuments as returned by Globus APIs.

    :ivar dict raw: The unparsed error subdocument
    """

    # the default set of fields to use for message extraction, in order
    # selected to match the fields defined by JSON:API by default
    DEFAULT_MESSAGE_FIELDS: tuple[str, ...] = ("detail", "title")

    def __init__(
        self, data: dict[str, t.Any], *, message_fields: t.Sequence[str] | None = None
    ) -> None:
        self.raw = data
        self._message_fields: tuple[str, ...]

        if message_fields is not None:
            self._message_fields = tuple(message_fields)
        else:
            self._message_fields = self.DEFAULT_MESSAGE_FIELDS

    @property
    def message(self) -> str | None:
        """
        The 'message' string of this subdocument, derived from its data based on the
        parsing context.

        May be `None` if no message is defined.
        """
        return _extract_message_from_dict(self.raw, self._message_fields)

    @property
    def code(self) -> str | None:
        """
        The 'code' string of this subdocument, derived from its data based on the
        parsing context.

        May be `None` if no code is defined.
        """
        if isinstance(self.raw.get("code"), str):
            return t.cast(str, self.raw["code"])
        return None

    def get(self, key: str, default: t.Any = None) -> t.Any | None:
        """
        A dict-like getter for the raw data.

        :param key: The string key to use for lookup
        :param default: The default value to use
        """
        return self.raw.get(key, default)


def _extract_message_from_dict(
    data: dict[str, t.Any], message_fields: tuple[str, ...]
) -> str | None:
    """
    Extract a single message string from a dict if one is present.
    """
    for f in message_fields:
        if isinstance(data.get(f), str):
            return t.cast(str, data[f])
    return None



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

INDEX_ID = "60d1160b-f016-40b0-8545-99619865873d"
IDENTITY_ID = "46bd0f56-e24f-11e5-a510-131bef46955c"
ROLE_ID = "MDQ1MzAy"


RESPONSES = ResponseSet(
    metadata={"index_id": INDEX_ID, "identity_id": IDENTITY_ID, "role_id": ROLE_ID},
    default=RegisteredResponse(
        service="search",
        path=f"/v1/index/{INDEX_ID}/role",
        method="POST",
        json={
            "creation_date": "2022-01-26 21:53:06",
            "id": ROLE_ID,
            "index_id": INDEX_ID,
            "principal": f"urn:globus:auth:identity:{IDENTITY_ID}",
            "principal_type": "identity",
            "role_name": "writer",
        },
        metadata={
            "index_id": INDEX_ID,
            "identity_id": IDENTITY_ID,
            "role_id": ROLE_ID,
            "role_name": "writer",
        },
    ),
)



import globus_sdk
from globus_sdk.testing import load_response


def test_get_function(compute_client_v2: globus_sdk.ComputeClientV2):
    meta = load_response(compute_client_v2.get_function).metadata
    res = compute_client_v2.get_function(function_id=meta["function_id"])
    assert res.http_status == 200
    assert res.data["function_uuid"] == meta["function_id"]
    assert res.data["function_name"] == meta["function_name"]
    assert res.data["function_code"] == meta["function_code"]



import json
import uuid

import pytest

from globus_sdk.testing import get_last_request, load_response


def test_get_endpoint(client):
    """
    Gets endpoint on fixture, validate results
    """
    meta = load_response(client.get_endpoint).metadata
    epid = meta["endpoint_id"]

    # load the endpoint document
    ep_doc = client.get_endpoint(epid)

    # check that the contents are basically OK
    assert ep_doc["DATA_TYPE"] == "endpoint"
    assert ep_doc["id"] == epid
    assert "display_name" in ep_doc


@pytest.mark.parametrize("epid_type", [uuid.UUID, str])
def test_update_endpoint(epid_type, client):
    meta = load_response(client.update_endpoint).metadata
    epid = meta["endpoint_id"]

    # NOTE: pass epid as UUID or str
    # requires that TransferClient correctly translates UUID
    update_data = {"display_name": "Updated Name", "description": "Updated description"}
    update_doc = client.update_endpoint(epid_type(epid), update_data)

    # make sure response is a successful update
    assert update_doc["DATA_TYPE"] == "result"
    assert update_doc["code"] == "Updated"
    assert update_doc["message"] == "Endpoint updated successfully"

    req = get_last_request()
    assert json.loads(req.body) == update_data



import uuid
from unittest import mock

import pytest
import requests

import globus_sdk


class MockDecoder(globus_sdk.IDTokenDecoder):
    def get_openid_configuration(self):
        return {
            "issuer": "https://auth.globus.org",
            "authorization_endpoint": "https://auth.globus.org/v2/oauth2/authorize",
            "userinfo_endpoint": "https://auth.globus.org/v2/oauth2/userinfo",
            "token_endpoint": "https://auth.globus.org/v2/oauth2/token",
            "revocation_endpoint": "https://auth.globus.org/v2/oauth2/token/revoke",
            "jwks_uri": "https://auth.globus.org/jwk.json",
            "response_types_supported": ["code", "token", "token id_token", "id_token"],
            "id_token_signing_alg_values_supported": ["RS512"],
            "scopes_supported": ["openid", "email", "profile"],
            "token_endpoint_auth_methods_supported": ["client_secret_basic"],
            "claims_supported": [
                "at_hash",
                "aud",
                "email",
                "exp",
                "name",
                "nonce",
                "preferred_username",
                "iat",
                "iss",
                "sub",
            ],
            "subject_types_supported": ["public"],
        }

    def get_jwk(self):
        return mock.Mock()


def test_decoding_defaults_to_client_id_as_audience():
    fake_client = mock.Mock()
    fake_client.client_id = str(uuid.uuid1())

    decoder = MockDecoder(fake_client)

    with mock.patch("jwt.decode") as mock_jwt_decode:
        decoder.decode("")
        assert mock_jwt_decode.call_args.kwargs["audience"] == fake_client.client_id


@pytest.mark.parametrize("audience_value", (None, "myaud"))
def test_decoding_passes_audience(audience_value):
    class MyDecoder(MockDecoder):
        def get_jwt_audience(self):
            return audience_value

    decoder = MyDecoder(mock.Mock())

    with mock.patch("jwt.decode") as mock_jwt_decode:
        decoder.decode("")
        assert mock_jwt_decode.call_args.kwargs["audience"] == audience_value


def test_setting_oidc_config_on_default_decoder_unpacks_data():
    oidc_config = {"x": 1}
    raw_response = mock.Mock(spec=requests.Response)
    raw_response.json.return_value = oidc_config
    response = globus_sdk.GlobusHTTPResponse(raw_response, client=mock.Mock())

    decoder = globus_sdk.IDTokenDecoder(mock.Mock())
    decoder.store_openid_configuration(response)

    assert decoder.get_openid_configuration() == oidc_config


def test_default_jwt_leeway_can_be_overridden_on_instance():
    decoder = MockDecoder(mock.Mock())
    default_leeway = decoder.jwt_leeway
    decoder.jwt_leeway = int(default_leeway * 2)

    with mock.patch("jwt.decode") as mock_jwt_decode:
        decoder.decode("")
        assert mock_jwt_decode.call_args.kwargs["leeway"] == int(default_leeway * 2)



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

GUARDIANS_IDP_ID = str(uuid.uuid1())
STAR_LORD = {
    "identity_provider": GUARDIANS_IDP_ID,
    "identity_type": "login",
    "organization": "Guardians of the Galaxy",
    "status": "used",
    "id": str(uuid.uuid1()),
    "name": "Star Lord",
    "username": "star.lord@guardians.galaxy",
    # I thought it would be funny if he didn't get 'star.lord'
    # because someone else got it first
    "email": "star.lord2@guardians.galaxy",
}
ROCKET_RACCOON = {
    "identity_provider": GUARDIANS_IDP_ID,
    "identity_type": "login",
    "organization": "Guardians of the Galaxy",
    "status": "used",
    "id": str(uuid.uuid1()),
    "name": "Rocket",
    "username": "rocket@guardians.galaxy",
    # and think about it, who else would try to lay claim
    # to that email address?
    "email": "star.lord@guardians.galaxy",
}

EVIL_TEST_PROJECT = {
    "admin_ids": [ROCKET_RACCOON["id"]],
    "contact_email": "eviltestproject@guardians.galaxy",
    "display_name": "Evil Test Project Full of Evil",
    "admin_group_ids": None,
    "id": str(uuid.uuid1()),
    "project_name": "Evil Test Project Full of Evil",
    "admins": {
        "identities": [ROCKET_RACCOON],
        "groups": [],
    },
}

GUARDIANS_PROJECT = {
    "admin_ids": [ROCKET_RACCOON["id"]],
    "contact_email": "support@guardians.galaxy",
    "display_name": "Guardians of the Galaxy Portal",
    "admin_group_ids": None,
    "id": str(uuid.uuid1()),
    "project_name": "Guardians of the Galaxy Portal",
    "admins": {
        "identities": [STAR_LORD, ROCKET_RACCOON],
        "groups": [],
    },
}

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="auth",
        path="/v2/api/projects",
        json={"projects": [EVIL_TEST_PROJECT, GUARDIANS_PROJECT]},
        metadata={
            "project_ids": [EVIL_TEST_PROJECT["id"], GUARDIANS_PROJECT["id"]],
            "admin_map": {
                EVIL_TEST_PROJECT["id"]: [ROCKET_RACCOON["id"]],
                GUARDIANS_PROJECT["id"]: [STAR_LORD["id"], ROCKET_RACCOON["id"]],
            },
        },
    )
)



from __future__ import annotations

import copy
import datetime
import typing as t
import uuid

from responses import matchers

from globus_sdk.testing import RegisteredResponse, ResponseList, ResponseSet

from ._common import RUN, RUN_ID, USER1

FLOW_ID_ALPHA = str(uuid.uuid1())
FLOW_ID_BETA = str(uuid.uuid1())


def generate_example_run(n: int, flow_id: str | None = None) -> dict[str, t.Any]:
    run_id = str(uuid.UUID(int=n))

    base_time = datetime.datetime.fromisoformat("2021-10-18T19:19:35.967289+00:00")
    start_time = base_time + datetime.timedelta(days=n)
    completion_time = base_time + datetime.timedelta(days=n + 1)

    run_doc = copy.deepcopy(RUN)
    run_doc["completion_time"] = completion_time.isoformat()
    run_doc["label"] = f"Run {n}"
    run_doc["start_time"] = start_time.isoformat()
    run_doc["run_id"] = run_id
    run_doc["action_id"] = run_id
    if flow_id is not None:
        run_doc["flow_id"] = flow_id

    return run_doc


_combined_filter_output = {
    "runs": [generate_example_run(i, flow_id=FLOW_ID_ALPHA) for i in range(5)]
    + [generate_example_run(i, flow_id=FLOW_ID_BETA) for i in range(5)],
    "limit": 10,
    "has_next_page": False,
}


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="flows",
        path="/runs",
        json={
            "runs": [RUN],
            "limit": 20,
            "has_next_page": False,
            "marker": None,
        },
        metadata={"first_run_id": RUN_ID},
    ),
    filter_flow_id=ResponseList(
        RegisteredResponse(
            service="flows",
            path="/runs",
            json={
                "runs": [
                    generate_example_run(i, flow_id=FLOW_ID_ALPHA) for i in range(5)
                ],
                "limit": 5,
                "has_next_page": False,
            },
            match=[matchers.query_param_matcher({"filter_flow_id": FLOW_ID_ALPHA})],
        ),
        RegisteredResponse(
            service="flows",
            path="/runs",
            json={
                "runs": [
                    generate_example_run(i, flow_id=FLOW_ID_BETA) for i in range(5)
                ],
                "limit": 5,
                "has_next_page": False,
            },
            match=[matchers.query_param_matcher({"filter_flow_id": FLOW_ID_BETA})],
        ),
        # register this twice to make the matching order insensitive
        RegisteredResponse(
            service="flows",
            path="/runs",
            json=_combined_filter_output,
            match=[
                matchers.query_param_matcher(
                    {"filter_flow_id": f"{FLOW_ID_ALPHA},{FLOW_ID_BETA}"}
                )
            ],
        ),
        RegisteredResponse(
            service="flows",
            path="/runs",
            json=_combined_filter_output,
            match=[
                matchers.query_param_matcher(
                    {"filter_flow_id": f"{FLOW_ID_BETA},{FLOW_ID_ALPHA}"}
                )
            ],
        ),
        metadata={
            "by_flow_id": {
                FLOW_ID_ALPHA: {
                    "num": 5,
                },
                FLOW_ID_BETA: {
                    "num": 5,
                },
            }
        },
    ),
    paginated=ResponseList(
        RegisteredResponse(
            service="flows",
            path="/runs",
            json={
                "runs": [generate_example_run(i) for i in range(20)],
                "limit": 20,
                "has_next_page": True,
                "marker": "fake_marker_0",
            },
        ),
        RegisteredResponse(
            service="flows",
            path="/runs",
            json={
                "runs": [generate_example_run(i) for i in range(20, 40)],
                "limit": 20,
                "has_next_page": True,
                "marker": "fake_marker_1",
            },
            match=[matchers.query_param_matcher({"marker": "fake_marker_0"})],
        ),
        RegisteredResponse(
            service="flows",
            path="/runs",
            json={
                "runs": [generate_example_run(i) for i in range(40, 60)],
                "limit": 20,
                "has_next_page": False,
                "marker": None,
            },
            match=[matchers.query_param_matcher({"marker": "fake_marker_1"})],
        ),
        metadata={
            "owner_id": USER1,
            "num_pages": 3,
            "expect_markers": ["fake_marker_0", "fake_marker_1", None],
            "total_items": 60,
        },
    ),
)



from __future__ import annotations

import abc
import contextlib
import os
import pathlib
import re
import sys
import typing as t
import uuid

import globus_sdk

from .token_data import TokenStorageData

if t.TYPE_CHECKING:
    from globus_sdk.globus_app import GlobusAppConfig


class TokenStorage(metaclass=abc.ABCMeta):
    """
    The interface for interacting with a store of :class:`TokenStorageData` objects.

    Implementations must partition their token data objects by ``namespace``.
    Within a namespace, token data must be indexed by ``resource_server``.

    :param namespace: A unique string for partitioning token data (Default: "DEFAULT").

    :ivar globus_sdk.IDTokenDecoder | None id_token_decoder: A decoder to use
        when decoding ``id_token`` JWTs from Globus Auth. By default, a new decoder
        is used each time decoding is performed.
    """

    def __init__(self, namespace: str = "DEFAULT") -> None:
        self.namespace = namespace
        self.id_token_decoder: globus_sdk.IDTokenDecoder | None = None

    @abc.abstractmethod
    def store_token_data_by_resource_server(
        self, token_data_by_resource_server: t.Mapping[str, TokenStorageData]
    ) -> None:
        """
        Store token data for one or more resource server in the current namespace.

        :param token_data_by_resource_server: mapping of resource server to token data.
        """

    @abc.abstractmethod
    def get_token_data_by_resource_server(self) -> dict[str, TokenStorageData]:
        """
        Retrieve all token data stored in the current namespace.

        :returns: a dict of ``TokenStorageData`` objects indexed by their
            resource server.
        """

    def get_token_data(self, resource_server: str) -> TokenStorageData | None:
        """
        Retrieve token data for a particular resource server in the current namespace.

        :param resource_server: The resource_server string to get token data for.
        :returns: token data if found or else None.
        """
        return self.get_token_data_by_resource_server().get(resource_server)

    @abc.abstractmethod
    def remove_token_data(self, resource_server: str) -> bool:
        """
        Remove token data for a resource server in the current namespace.

        :param resource_server: The resource server string to remove token data for.
        :returns: True if token data was deleted, False if none was found to delete.
        """

    def store_token_response(
        self, token_response: globus_sdk.OAuthTokenResponse
    ) -> None:
        """
        Store token data from an :class:`OAuthTokenResponse` in the current namespace.

        :param token_response: A token response object from an authentication flow.
        """
        token_data_by_resource_server = {}

        identity_id = self._extract_identity_id(token_response)

        for resource_server, token_dict in token_response.by_resource_server.items():
            token_data_by_resource_server[resource_server] = TokenStorageData(
                resource_server=token_dict["resource_server"],
                identity_id=identity_id,
                scope=token_dict["scope"],
                access_token=token_dict["access_token"],
                refresh_token=token_dict.get("refresh_token"),
                expires_at_seconds=token_dict["expires_at_seconds"],
                token_type=token_dict.get("token_type"),
            )
        self.store_token_data_by_resource_server(token_data_by_resource_server)

    def _extract_identity_id(
        self, token_response: globus_sdk.OAuthTokenResponse
    ) -> str | None:
        """
        Get identity_id from id_token if available.

        .. note::

            This method is private, but is used in ValidatingTokenStorage to
            override the extraction of ``identity_id`` information.

        Generalizing customization of ``identity_id`` extraction will require
        implementation of a user-facing mechanism for controlling calls to
        ``decode_id_token()``.
        """
        # dependent token responses cannot contain an `id_token` field, as the
        # top-level data is an array
        if isinstance(token_response, globus_sdk.OAuthDependentTokenResponse):
            return None

        if id_token := token_response.get("id_token"):
            if self.id_token_decoder:
                decoded_id_token = self.id_token_decoder.decode(id_token)
            else:
                decoded_id_token = token_response.decode_id_token()
            return decoded_id_token["sub"]  # type: ignore[no-any-return]
        else:
            return None

    def close(self) -> None:  # noqa: B027
        """
        Close any resources associated with this token storage.

        By default, this does nothing, but subclasses may override it.
        """


class FileTokenStorage(TokenStorage, metaclass=abc.ABCMeta):
    """
    A base class for token storages which store tokens in a local file.

    Common functionality for file-based token storages like file creation and class
    instantiation for a GlobusApp is defined here.

    :cvar file_format: The file format suffix associated with files of this type. This
        is used when constructing the file path for a GlobusApp.

    :param filepath: The path to a file where token data should be stored.
    :param namespace: A unique string for partitioning token data (Default: "DEFAULT").
    """

    # File suffix associated with files of this type (e.g., "csv")
    file_format: str = "_UNSET_"  # must be overridden by subclasses

    def __init__(
        self, filepath: pathlib.Path | str, *, namespace: str = "DEFAULT"
    ) -> None:
        """
        :param filepath: the name of the file to write to and read from
        :param namespace: A user-supplied namespace for partitioning token data
        """
        self.filepath = str(filepath)
        try:
            self._ensure_containing_dir_exists()
        except OSError as e:
            msg = (
                "Encountered an error while initializing the token storage file "
                f"'{self.filepath}'"
            )
            raise ValueError(msg) from e
        super().__init__(namespace=namespace)

    def __init_subclass__(cls, **kwargs: t.Any):
        if cls.file_format == "_UNSET_":
            raise TypeError(f"{cls.__name__} must set a 'file_format' class attribute")

    @classmethod
    def for_globus_app(
        cls,
        client_id: uuid.UUID | str,
        app_name: str,
        config: GlobusAppConfig,
        namespace: str,
    ) -> TokenStorage:
        """
        Initialize a TokenStorage instance for a GlobusApp, using the supplied
        info to determine the file location.

        :param client_id: The client ID of the Globus App.
        :param app_name: The name of the Globus App.
        :param config: The GlobusAppConfig object for the Globus App.
        :param namespace: A user-supplied namespace for partitioning token data.
        """
        filepath = _default_globus_app_filepath(client_id, app_name, config.environment)
        return cls(filepath=f"{filepath}.{cls.file_format}", namespace=namespace)

    def _ensure_containing_dir_exists(self) -> None:
        """
        Ensure that the directory containing the given filepath exists.
        """
        os.makedirs(os.path.dirname(self.filepath), exist_ok=True)

    def file_exists(self) -> bool:
        """
        Check if the file used by this file storage adapter exists.
        """
        return os.path.exists(self.filepath)

    @contextlib.contextmanager
    def user_only_umask(self) -> t.Iterator[None]:
        """
        A context manager to deny rwx to Group and World, x to User

        This does not create a file, but ensures that if a file is created while in the
        context manager, its permissions will be correct on unix systems.

        .. note::

            On Windows, this has no effect. To control the permissions on files used for
            token storage, use ``%LOCALAPPDATA%`` or ``%APPDATA%``.
            These directories should only be accessible to the current user.
        """
        old_umask = os.umask(0o177)
        try:
            yield
        finally:
            os.umask(old_umask)


def _default_globus_app_filepath(
    client_id: uuid.UUID | str, app_name: str, environment: str
) -> str:
    r"""
    Construct a default TokenStorage filepath for a GlobusApp.
    For flexibility, the filepath will omit the file format suffix.

    On Windows, this will be:
        ``~\AppData\Local\globus\app\{client_id}\{app_name}\tokens``

    On Linux and macOS, we use:
        ``~/.globus/app/{client_id}/{app_name}/tokens``
    """
    environment_prefix = f"{environment}-"
    if environment == "production":
        environment_prefix = ""
    filename = f"{environment_prefix}tokens"
    app_name = _slugify_app_name(app_name)

    if sys.platform == "win32":
        # try to get the app data dir, preferring the local appdata
        datadir = os.getenv("LOCALAPPDATA", os.getenv("APPDATA"))
        if not datadir:
            home = os.path.expanduser("~")
            datadir = os.path.join(home, "AppData", "Local")
        return os.path.join(
            datadir, "globus", "app", str(client_id), app_name, filename
        )
    else:
        return os.path.expanduser(
            f"~/.globus/app/{str(client_id)}/{app_name}/{filename}"
        )


# https://learn.microsoft.com/en-us/windows/win32/fileio/naming-a-file#naming-conventions
_RESERVED_FS_CHARS = re.compile(r'[<>:"/\\|?*]')
# https://stackoverflow.com/a/31976060
_RESERVED_FS_NAMES = re.compile(r"con|prn|aux|nul|com\d|lpt\d")


def _slugify_app_name(app_name: str) -> str:
    """
    Slugify a globus app name for use in a file path.

    * Reserved filesystem characters are replaced with a '+'. ('a?' -> 'a+')
    * Periods and Spaces are replaced with a '-'. ('a. b' -> 'a--b')
    * Control characters are removed. ('a\0b' -> 'ab')
    * The string is lowercased. ('AB' -> 'ab')

    :raises: GlobusSDKUsageError if the app name is empty after slugification.
    :raises: GlobusSDKUsageError if the app name is a reserved filesystem name (after
        slugification).
    """
    app_name = _RESERVED_FS_CHARS.sub("+", app_name)
    app_name = app_name.replace(".", "-").replace(" ", "-")
    app_name = "".join(c for c in app_name if c.isprintable())
    app_name = app_name.lower()

    if _RESERVED_FS_NAMES.fullmatch(app_name):
        msg = (
            f'App name results in a reserved filename ("{app_name}"). '
            "Please choose a different name."
        )
        raise globus_sdk.GlobusSDKUsageError(msg)

    if not app_name:
        msg = "App name results in the empty string. Please choose a different name."
        raise globus_sdk.GlobusSDKUsageError(msg)

    return app_name



import urllib.parse

import globus_sdk
from globus_sdk.testing import get_last_request, load_response


def test_get_endpoints(compute_client_v2: globus_sdk.ComputeClientV2):
    meta = load_response(compute_client_v2.get_endpoints).metadata

    res = compute_client_v2.get_endpoints()

    assert res.http_status == 200
    assert res.data[0]["uuid"] == meta["endpoint_id"]
    assert res.data[1]["uuid"] == meta["endpoint_id_2"]

    # confirm that the 'role' param was not sent
    last_req = get_last_request()
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(last_req.url).query)
    assert "role" not in parsed_qs


def test_get_endpoints_any(compute_client_v2: globus_sdk.ComputeClientV2):
    meta = load_response(compute_client_v2.get_endpoints, case="any").metadata
    res = compute_client_v2.get_endpoints(role="any")

    last_req = get_last_request()
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(last_req.url).query)
    assert parsed_qs["role"] == ["any"]

    assert res.http_status == 200
    assert res.data[0]["uuid"] == meta["endpoint_id"]
    assert res.data[1]["uuid"] == meta["endpoint_id_2"]
    assert res.data[2]["uuid"] == meta["endpoint_id_3"]
    assert res.data[1]["owner"] != res.data[2]["owner"]



import pytest

import globus_sdk
from globus_sdk.testing import load_response


def test_set_subscription_id(client):
    meta = load_response(client.set_subscription_id).metadata
    epid = meta["endpoint_id"]

    res = client.set_subscription_id(epid, "DEFAULT")
    assert res["code"] == "Updated"
    assert res["message"] == "Endpoint updated successfully"


def test_set_subscription_id_fails_notfound(client):
    meta = load_response(client.set_subscription_id, case="not_found").metadata
    epid = meta["endpoint_id"]

    with pytest.raises(globus_sdk.TransferAPIError) as excinfo:
        client.set_subscription_id(epid, "DEFAULT")

    assert excinfo.value.code == "EndpointNotFound"
    assert excinfo.value.messages == [f"No such endpoint '{epid}'"]


def test_set_subscription_id_fails_multi(client):
    meta = load_response(
        client.set_subscription_id, case="multi_subscriber_cannot_use_default"
    ).metadata
    epid = meta["endpoint_id"]

    with pytest.raises(globus_sdk.TransferAPIError) as excinfo:
        client.set_subscription_id(epid, "DEFAULT")

    assert excinfo.value.code == "BadRequest"
    assert len(excinfo.value.messages) == 1
    message = excinfo.value.messages[0]
    assert message.startswith("Please specify the subscription ID to use.")



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

endpoint_id = str(uuid.uuid4())
gcs_manager_url = RegisteredResponse._url_map["gcs"]
display_name = "Happy Fun Endpoint"

RESPONSES = ResponseSet(
    metadata={
        "endpoint_id": endpoint_id,
        "display_name": display_name,
        "gcs_manager_url": gcs_manager_url,
    },
    default=RegisteredResponse(
        service="gcs",
        path="/endpoint",
        json={
            "DATA_TYPE": "result#1.0.0",
            "code": "success",
            "data": [
                {
                    "DATA_TYPE": "endpoint#1.2.0",
                    "allow_udt": False,
                    "contact_email": "user@globus.org",
                    "display_name": display_name,
                    "gcs_manager_url": gcs_manager_url,
                    "gridftp_control_channel_port": 443,
                    "id": endpoint_id,
                    "network_use": "normal",
                    "organization": "Globus",
                    "public": True,
                    "subscription_id": None,
                }
            ],
            "detail": "success",
            "has_next_page": False,
            "http_response_code": 200,
        },
    ),
)



from __future__ import annotations

import typing as t
import uuid

from ..collection import (
    DynamicScopeCollection,
    StaticScopeCollection,
    _url_scope,
)
from ..representation import Scope


class _FlowsScopes(StaticScopeCollection):
    # The Flows service breaks the scopes/resource server convention: its
    # resource server is a domain name but its scopes are built around the
    # client ID.
    resource_server = "flows.globus.org"
    client_id = "eec9b274-0c81-4334-bdc2-54e90e689b9a"

    all = _url_scope(client_id, "all")
    manage_flows = _url_scope(client_id, "manage_flows")
    view_flows = _url_scope(client_id, "view_flows")
    run = _url_scope(client_id, "run")
    run_status = _url_scope(client_id, "run_status")
    run_manage = _url_scope(client_id, "run_manage")


FlowsScopes = _FlowsScopes()


class SpecificFlowScopes(DynamicScopeCollection):
    """
    This defines the scopes for a single flow (as distinct from the Flows service).

    It primarily provides the `user` scope which is typically needed to start a run of
    a flow.

    Example usage:

    .. code-block:: python

        sc = SpecificFlowScopes("my-flow-id-here")
        flow_scope = sc.user
    """

    _scope_names = ("user",)

    def __init__(self, flow_id: uuid.UUID | str) -> None:
        _flow_id = str(flow_id)
        super().__init__(_flow_id)

        self.user: Scope = _url_scope(
            _flow_id, f"flow_{_flow_id.replace('-', '_')}_user"
        )

    @classmethod
    def _build_class_stub(cls) -> SpecificFlowScopes:
        """
        This internal helper builds a "stub" object so that
        ``SpecificFlowClient.scopes`` is typed as ``SpecificFlowScopes`` but
        raises appropriate errors at runtime access via the class.
        """
        return _SpecificFlowScopesClassStub()


class _SpecificFlowScopesClassStub(SpecificFlowScopes):
    """
    This stub object ensures that the type deductions for type checkers (e.g. mypy) on
    SpecificFlowClient.scopes are correct.

    Primarily, it should be possible to access the `scopes` attribute, the `user`
    scope, and the `resource_server`, but these usages should raise specific and
    informative runtime errors.

    Our types are therefore less accurate for class-var access, but more accurate for
    instance-var access.
    """

    def __init__(self) -> None:
        super().__init__("<stub>")

    def __getattribute__(self, name: str) -> t.Any:
        if name == "user":
            _raise_attr_error("scopes")
        if name == "resource_server":
            _raise_attr_error("resource_server")
        return object.__getattribute__(self, name)


def _raise_attr_error(name: str) -> t.NoReturn:
    raise AttributeError(
        f"It is not valid to attempt to access the '{name}' attribute of the "
        "SpecificFlowClient class. "
        f"Instead, instantiate a SpecificFlowClient and access the '{name}' attribute "
        "from that instance."
    )



import typing as t

from .add_content_directive import AddContentDirective


class PaginatedUsage(AddContentDirective):
    has_content = False
    required_arguments = 1
    optional_arguments = 0

    def gen_rst(self) -> t.Iterator[str]:
        yield "This method supports paginated access. "
        yield "To use the paginated variant, give the same arguments as normal, "
        yield "but prefix the method name with ``paginated``, as in"
        yield ""
        yield ".. code-block::"
        yield ""
        yield f"    client.paginated.{self.arguments[0]}(...)"
        yield ""
        yield "For more information, see"
        yield ":ref:`how to make paginated calls <making_paginated_calls>`."



import pytest

import globus_sdk
from globus_sdk.testing import load_response


def test_create_index(client):
    meta = load_response(client.create_index).metadata

    res = client.create_index("Foo Title", "bar description")
    assert res.http_status == 200
    assert res["id"] == meta["index_id"]


def test_create_index_limit_exceeded(client):
    load_response(client.create_index, case="trial_limit")

    with pytest.raises(globus_sdk.SearchAPIError) as excinfo:
        client.create_index("Foo Title", "bar description")

    err = excinfo.value

    assert err.http_status == 409
    assert err.code == "Conflict.LimitExceeded"



from globus_sdk import GCSClient
from globus_sdk.testing import load_response


def test_client_address_handling():
    # variants of the same location
    c1 = GCSClient("foo.data.globus.org")
    c2 = GCSClient("https://foo.data.globus.org")
    c3 = GCSClient("https://foo.data.globus.org/api/")
    # explicit subpath of /api/
    c4 = GCSClient("https://foo.data.globus.org/api/bar")
    # explicit construction can point at the root (rather than /api/)
    c5 = GCSClient("foo.data.globus.org")
    c5.base_url = "https://foo.data.globus.org/"

    # 1, 2, and 3 are all the same
    assert c1.base_url == c2.base_url
    assert c1.base_url == c3.base_url

    # 4 and 5 are different from the rest
    assert c4.base_url != c1.base_url
    assert c5.base_url != c1.base_url

    # 5 is the root of 1
    assert c1.base_url.startswith(c5.base_url)


def test_gcs_client_resource_server_and_endpoint_client_id():
    meta = load_response(GCSClient.get_gcs_info).metadata
    endpoint_client_id = meta["endpoint_client_id"]
    domain_name = meta["domain_name"]

    c = GCSClient(domain_name)
    assert c.endpoint_client_id == endpoint_client_id
    assert c.resource_server == endpoint_client_id



from ..collection import StaticScopeCollection, _url_scope


class _TimersScopes(StaticScopeCollection):
    resource_server = "524230d7-ea86-4a52-8312-86065a9e0417"

    timer = _url_scope(resource_server, "timer")


TimersScopes = _TimersScopes()



from __future__ import annotations

import logging
import typing as t

import globus_sdk

from .renewing import RenewingAuthorizer

log = logging.getLogger(__name__)


class RefreshTokenAuthorizer(
    RenewingAuthorizer["globus_sdk.OAuthRefreshTokenResponse"]
):
    """
    Implements Authorization using a Refresh Token to periodically fetch
    renewed Access Tokens. It may be initialized with an Access Token, or it
    will fetch one the first time that ``get_authorization_header()`` is
    called.

    Example usage looks something like this:

    .. code-block:: pycon

        >>> import globus_sdk
        >>> auth_client = globus_sdk.ConfidentialAppAuthClient(client_id=..., client_secret=...)
        >>> # do some flow to get a refresh token from auth_client
        >>> rt_authorizer = globus_sdk.RefreshTokenAuthorizer(refresh_token, auth_client)
        >>> # create a new client
        >>> transfer_client = globus_sdk.TransferClient(authorizer=rt_authorizer)

    Anything which inherits from :class:`BaseClient <globus_sdk.BaseClient>`
    will automatically support usage of the ``RefreshTokenAuthorizer``.

    :param refresh_token: Refresh Token for Globus Auth
    :param auth_client: ``AuthClient`` capable of using the ``refresh_token``
    :param access_token: Initial Access Token to use, only used if ``expires_at`` is
        also set
    :param expires_at: Expiration time for the starting ``access_token`` expressed as a
        POSIX timestamp (i.e. seconds since the epoch)
    :param on_refresh: A callback which is triggered any time this authorizer fetches a
        new access_token. The ``on_refresh`` callable is invoked on the
        :class:`globus_sdk.OAuthRefreshTokenResponse` object resulting from the token being
        refreshed. It should take only one argument, the token response object.
        This is useful for implementing storage for Access Tokens, as the
        ``on_refresh`` callback can be used to update the Access Tokens and
        their expiration times.
    """  # noqa: E501

    def __init__(
        self,
        refresh_token: str,
        auth_client: globus_sdk.AuthLoginClient,
        *,
        access_token: str | None = None,
        expires_at: int | None = None,
        on_refresh: (
            None | t.Callable[[globus_sdk.OAuthRefreshTokenResponse], t.Any]
        ) = None,
    ) -> None:
        log.debug(
            "Setting up RefreshTokenAuthorizer with auth_client="
            f"[instance:{id(auth_client)}]"
        )
        # per type checkers, this is unreachable... but it is, of course, reachable...
        # that's... the point
        if isinstance(auth_client, globus_sdk.AuthClient):  # type: ignore[unreachable]
            raise globus_sdk.GlobusSDKUsageError(
                "RefreshTokenAuthorizer requires an AuthLoginClient, not an "
                "AuthClient. In past versions of the SDK, it was possible to "
                "use an AuthClient if it was correctly authorized, but this is "
                "no longer allowed. "
                "Proper usage should typically use a NativeAppAuthClient or "
                "ConfidentialAppAuthClient."
            )

        # required for _get_token_data
        self.refresh_token = refresh_token
        self.auth_client = auth_client

        super().__init__(access_token, expires_at, on_refresh)

    def _get_token_response(self) -> globus_sdk.OAuthRefreshTokenResponse:
        """
        Make a refresh token grant
        """
        return self.auth_client.oauth2_refresh_token(self.refresh_token)

    def _extract_token_data(
        self, res: globus_sdk.OAuthRefreshTokenResponse
    ) -> dict[str, t.Any]:
        """
        Get the tokens .by_resource_server,
        Ensure that only one token was gotten, and return that token.

        If the token_data includes a "refresh_token" field, update self.refresh_token to
        that value.
        """
        token_data_list = list(res.by_resource_server.values())
        if len(token_data_list) != 1:
            raise ValueError(
                "Attempting refresh for refresh token authorizer "
                "didn't return exactly one token. Possible service error."
            )

        token_data = next(iter(token_data_list))

        # handle refresh_token being present
        # mandated by OAuth2: https://tools.ietf.org/html/rfc6749#section-6
        if "refresh_token" in token_data:
            self.refresh_token = token_data["refresh_token"]

        return token_data



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import TIMER_ID, V1_TIMER

TIMER_INACTIVE_USER = {
    **V1_TIMER,
    "status": "inactive",
    "inactive_reason": {"cause": "user", "detail": None},
}

TIMER_INACTIVE_GARE = {
    **V1_TIMER,
    "status": "inactive",
    "inactive_reason": {
        "cause": "globus_auth_requirements",
        "detail": {
            "code": "ConsentRequired",
            "authorization_parameters": {
                "session_message": "Missing required data_access consent",
                "required_scopes": [
                    (
                        "https://auth.globus.org/scopes/actions.globus.org/"
                        "transfer/transfer"
                        "[urn:globus:auth:scope:transfer.api.globus.org:all"
                        "[*https://auth.globus.org/scopes/"
                        "543aade1-db97-4a4b-9bdf-0b58e78dfa69/data_access]]"
                    )
                ],
            },
        },
    },
}


RESPONSES = ResponseSet(
    metadata={"job_id": TIMER_ID},
    default=RegisteredResponse(
        service="timer",
        path=f"/jobs/{TIMER_ID}",
        method="GET",
        json=V1_TIMER,
    ),
    inactive_gare=RegisteredResponse(
        service="timer",
        path=f"/jobs/{TIMER_ID}",
        method="GET",
        json=TIMER_INACTIVE_GARE,
    ),
    inactive_user=RegisteredResponse(
        service="timer",
        path=f"/jobs/{TIMER_ID}",
        method="GET",
        json=TIMER_INACTIVE_USER,
    ),
    simple_500_error=RegisteredResponse(
        service="timer",
        path=f"/jobs/{TIMER_ID}",
        method="GET",
        status=500,
        json={
            "error": {
                "code": "ERROR",
                "detail": "Request failed terribly",
                "status": 500,
            }
        },
    ),
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import FORBIDDEN_AUTH_RESPONSE, UNAUTHORIZED_AUTH_RESPONSE

RESPONSES = ResponseSet(
    unauthorized=RegisteredResponse(
        service="auth",
        path="/v2/oauth2/userinfo",
        status=401,
        json=UNAUTHORIZED_AUTH_RESPONSE.json,
        metadata={
            "error_id": UNAUTHORIZED_AUTH_RESPONSE.error_id,
            **UNAUTHORIZED_AUTH_RESPONSE.metadata_include,
        },
    ),
    forbidden=RegisteredResponse(
        service="auth",
        path="/v2/oauth2/userinfo",
        status=403,
        json=FORBIDDEN_AUTH_RESPONSE.json,
        metadata={
            "error_id": FORBIDDEN_AUTH_RESPONSE.error_id,
            **FORBIDDEN_AUTH_RESPONSE.metadata_include,
        },
    ),
)



import typing as t

from docutils import nodes
from docutils.nodes import Node
from docutils.parsers.rst import Directive
from docutils.statemachine import StringList
from sphinx.util.nodes import nested_parse_with_titles


class AddContentDirective(Directive):
    # Directive source:
    #   https://sourceforge.net/p/docutils/code/HEAD/tree/trunk/docutils/docutils/parsers/rst/__init__.py
    #
    # for information on how to write directives, see also:
    #   https://docutils.sourceforge.io/docs/howto/rst-directives.html#the-directive-class
    def gen_rst(self) -> t.Iterator[str]:
        yield from []

    def run(self) -> t.Sequence[Node]:
        # implementation is based on the docs for parsing directive content as ReST doc:
        #   https://www.sphinx-doc.org/en/master/extdev/markupapi.html#parsing-directive-content-as-rest
        # which is how we will create rst here and have sphinx parse it

        # we need to add content lines into a ViewList which represents the content
        # of this directive, each one is appended with a "source" name
        # the best documentation for docutils is the source, so for ViewList, see...
        #   https://sourceforge.net/p/docutils/code/HEAD/tree/trunk/docutils/docutils/statemachine.py
        # writing the source name with angle-brackets seems to be the norm, but it's not
        # clear why -- perhaps docutils will care about the source in some way
        viewlist = StringList()
        linemarker = "<" + self.__class__.__name__.lower() + ">"
        for line in self.gen_rst():
            viewlist.append(line, linemarker)

        # create a section node (it doesn't really matter what node type we use here)
        # and add the viewlist we just created as the children of the new node via the
        # nested_parse method
        # then return the children (i.e. discard the docutils node)
        node = nodes.section()
        node.document = self.state.document
        nested_parse_with_titles(self.state, viewlist, node)
        return node.children



from globus_sdk import TransferAPIError
from globus_sdk.testing import construct_error


def test_transfer_response_get_args():
    err = construct_error(
        error_class=TransferAPIError,
        body={
            "message": "transfer error message",
            "code": "Transfer Error",
            "request_id": "123",
        },
        http_status=404,
    )
    req = err._underlying_response.request
    assert err._get_args() == [
        req.method,
        req.url,
        None,
        404,
        "Transfer Error",
        "transfer error message",
        "123",
    ]



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import BASE_GROUP_DOC, GROUP_ID

RESPONSES = ResponseSet(
    metadata={"group_id": GROUP_ID},
    default=RegisteredResponse(
        service="groups",
        path=f"/v2/groups/{GROUP_ID}",
        method="DELETE",
        json=BASE_GROUP_DOC,
    ),
)



from __future__ import annotations

import contextlib
import logging
import pathlib
import time
import typing as t

import requests

from globus_sdk import __version__, config, exc
from globus_sdk.authorizers import GlobusAuthorizer
from globus_sdk.transport.encoders import (
    FormRequestEncoder,
    JSONRequestEncoder,
    RequestEncoder,
)

from ._clientinfo import GlobusClientInfo
from .caller_info import RequestCallerInfo
from .retry import RetryContext
from .retry_check_runner import RetryCheckRunner
from .retry_config import RetryConfig

log = logging.getLogger(__name__)


class RequestsTransport:
    """
    The RequestsTransport handles HTTP request sending and retries.

    It receives raw request information from a client class, and then performs the
    following steps
    - encode the data in a prepared request
    - repeatedly send the request until no retry is requested by the configured hooks
    - return the last response or reraise the last exception

    If the maximum number of retries is reached, the final response or exception will
    be returned or raised.

    :param verify_ssl: Explicitly enable or disable SSL verification,
        or configure the path to a CA certificate bundle to use for SSL verification
    :param http_timeout: Explicitly set an HTTP timeout value in seconds. This parameter
        defaults to 60s but can be set via the ``GLOBUS_SDK_HTTP_TIMEOUT`` environment
        variable. Any value set via this parameter takes precedence over the environment
        variable.

    :ivar dict[str, str] headers: The headers which are sent on every request. These
        may be augmented by the transport when sending requests.
    """

    #: default maximum number of retries
    DEFAULT_MAX_RETRIES = 5

    #: the encoders are a mapping of encoding names to encoder objects
    encoders: dict[str, RequestEncoder] = {
        "text": RequestEncoder(),
        "json": JSONRequestEncoder(),
        "form": FormRequestEncoder(),
    }

    BASE_USER_AGENT = f"globus-sdk-py-{__version__}"

    def __init__(
        self,
        verify_ssl: bool | str | pathlib.Path | None = None,
        http_timeout: float | None = None,
    ) -> None:
        self.session = requests.Session()
        self.verify_ssl = config.get_ssl_verify(verify_ssl)
        self.http_timeout = config.get_http_timeout(http_timeout)
        self._user_agent = self.BASE_USER_AGENT
        self.globus_client_info: GlobusClientInfo = GlobusClientInfo(
            update_callback=self._handle_clientinfo_update
        )
        self.headers: dict[str, str] = {
            "Accept": "application/json",
            "User-Agent": self.user_agent,
            "X-Globus-Client-Info": self.globus_client_info.format(),
        }

    def close(self) -> None:
        """
        Closes all resources owned by the transport, primarily the underlying
        network session.
        """
        self.session.close()

    @property
    def user_agent(self) -> str:
        return self._user_agent

    @user_agent.setter
    def user_agent(self, value: str) -> None:
        """
        Set the ``user_agent`` and update the ``User-Agent`` header in ``headers``.

        :param value: The new user-agent string to set (after the base user-agent)
        """
        self._user_agent = f"{self.BASE_USER_AGENT}/{value}"
        self.headers["User-Agent"] = self._user_agent

    def _handle_clientinfo_update(
        self,
        info: GlobusClientInfo,  # pylint: disable=unused-argument
    ) -> None:
        """
        When the attached ``GlobusClientInfo`` is updated, write it back into
        ``headers``.

        If the client info is cleared, it will be removed from the headers.
        """
        formatted = self.globus_client_info.format()
        if formatted:
            self.headers["X-Globus-Client-Info"] = formatted
        else:
            # discard the element, so that this can be invoked multiple times
            self.headers.pop("X-Globus-Client-Info", None)

    @contextlib.contextmanager
    def tune(
        self,
        *,
        verify_ssl: bool | str | pathlib.Path | None = None,
        http_timeout: float | None = None,
    ) -> t.Iterator[None]:
        """
        Temporarily adjust some of the request sending settings of the transport.
        This method works as a context manager, and will reset settings to their
        original values after it exits.

        :param verify_ssl: Explicitly enable or disable SSL verification,
            or configure the path to a CA certificate bundle to use for SSL verification
        :param http_timeout: Explicitly set an HTTP timeout value in seconds

        **Example Usage**

        This can be used with any client class to temporarily set values in the context
        of one or more HTTP requests. To increase the HTTP request timeout from the
        default of 60 to 120 seconds,

        >>> client = ...  # any client class
        >>> with client.transport.tune(http_timeout=120):
        >>>     foo = client.get_foo()

        See also: :meth:`RetryConfig.tune`.
        """
        saved_settings = (
            self.verify_ssl,
            self.http_timeout,
        )
        if verify_ssl is not None:
            if isinstance(verify_ssl, bool):
                self.verify_ssl = verify_ssl
            else:
                self.verify_ssl = str(verify_ssl)
        if http_timeout is not None:
            self.http_timeout = http_timeout
        yield
        (
            self.verify_ssl,
            self.http_timeout,
        ) = saved_settings

    def _encode(
        self,
        method: str,
        url: str,
        query_params: dict[str, t.Any] | None = None,
        data: dict[str, t.Any] | list[t.Any] | str | bytes | None = None,
        headers: dict[str, str] | None = None,
        encoding: str | None = None,
    ) -> requests.Request:
        if headers:
            headers = {**self.headers, **headers}
        else:
            headers = self.headers

        if encoding is None:
            if isinstance(data, (bytes, str)):
                encoding = "text"
            else:
                encoding = "json"

        if encoding not in self.encoders:
            raise ValueError(
                f"Unknown encoding '{encoding}' is not supported by this transport."
            )

        return self.encoders[encoding].encode(method, url, query_params, data, headers)

    def _set_authz_header(
        self, authorizer: GlobusAuthorizer | None, req: requests.Request
    ) -> None:
        if authorizer:
            authz_header = authorizer.get_authorization_header()
            if authz_header:
                req.headers["Authorization"] = authz_header
            else:
                req.headers.pop("Authorization", None)  # remove any possible value

    def _retry_sleep(self, retry_config: RetryConfig, ctx: RetryContext) -> None:
        """
        Given a retry context, compute the amount of time to sleep and sleep that much
        This is always the minimum of the backoff (run on the context) and the
        ``max_sleep``.

        :param ctx: The context object which describes the state of the request and the
            retries which may already have been attempted.
        """
        sleep_period = min(retry_config.backoff(ctx), retry_config.max_sleep)
        log.debug(
            "request retry_sleep(%s) [max=%s]",
            sleep_period,
            retry_config.max_sleep,
        )
        time.sleep(sleep_period)

    def request(
        self,
        method: str,
        url: str,
        *,
        caller_info: RequestCallerInfo,
        query_params: dict[str, t.Any] | None = None,
        data: dict[str, t.Any] | list[t.Any] | str | bytes | None = None,
        headers: dict[str, str] | None = None,
        encoding: str | None = None,
        allow_redirects: bool = True,
        stream: bool = False,
    ) -> requests.Response:
        """
        Send an HTTP request

        :param url: URL for the request
        :param method: HTTP request method, as an all caps string
        :param caller_info: Contextual information about the caller of the request,
            including the authorizer and retry configuration.
        :param query_params: Parameters to be encoded as a query string
        :param headers: HTTP headers to add to the request
        :param data: Data to send as the request body. May pass through encoding.
        :param encoding: A way to encode request data. "json", "form", and "text"
            are all valid values. Custom encodings can be used only if they are
            registered with the transport. By default, strings get "text" behavior and
            all other objects get "json".
        :param allow_redirects: Follow Location headers on redirect response
            automatically. Defaults to ``True``
        :param stream: Do not immediately download the response content. Defaults to
            ``False``

        :return: ``requests.Response`` object
        """
        log.debug("starting request for %s", url)
        resp: requests.Response | None = None
        req = self._encode(method, url, query_params, data, headers, encoding)
        retry_config = caller_info.retry_config
        checker = RetryCheckRunner(caller_info.retry_config.checks)

        log.debug("transport request state initialized")
        for attempt in range(retry_config.max_retries + 1):
            log.debug("transport request retry cycle. attempt=%d", attempt)
            # add Authorization header, or (if it's a NullAuthorizer) possibly
            # explicitly remove the Authorization header
            # done fresh for each request, to handle potential for refreshed credentials
            self._set_authz_header(caller_info.authorizer, req)

            ctx = RetryContext(attempt, caller_info=caller_info)
            try:
                log.debug("request about to send")
                resp = ctx.response = self.session.send(
                    req.prepare(),
                    timeout=self.http_timeout,
                    verify=self.verify_ssl,
                    allow_redirects=allow_redirects,
                    stream=stream,
                )
            except requests.RequestException as err:
                log.debug("request hit error (RequestException)")
                ctx.exception = err
                if attempt >= retry_config.max_retries or not checker.should_retry(ctx):
                    log.warning("request done (fail, error)")
                    raise exc.convert_request_exception(err)
                log.debug("request may retry (should-retry=true)")
            else:
                log.debug("request success, still check should-retry")
                if not checker.should_retry(ctx):
                    log.debug("request done (success)")
                    return resp
                log.debug("request may retry, will check attempts")

            # the request will be retried, so sleep...
            if attempt < retry_config.max_retries:
                log.debug("under attempt limit, will sleep")
                self._retry_sleep(retry_config, ctx)
        if resp is None:
            raise ValueError("Somehow, retries ended without a response")
        log.warning("request reached max retries, done (fail, response)")
        return resp



import globus_sdk
from globus_sdk.testing import load_response


def test_get_task_group(compute_client_v2: globus_sdk.ComputeClientV2):
    meta = load_response(compute_client_v2.get_task_group).metadata
    res = compute_client_v2.get_task_group(task_group_id=meta["task_group_id"])
    assert res.http_status == 200
    assert meta["task_group_id"] == res.data["taskgroup_id"]
    assert meta["task_id"] == res.data["tasks"][0]["id"]
    assert meta["task_id_2"] == res.data["tasks"][1]["id"]



"""
This module provides convenience data structures to model and interact with
    Globus Auth consents.

The resources defined herein are:
    * ``Consent``
        * A data object modeling a user's grant for a client to perform some scoped
            operation on their behalf.
        * This grant is conditional on the invocation path leading to the client's
            attempted operation being initiated through a chain of similarly scoped
            operations (consents) defined in the "dependency_path".
    * ``ConsentTree``
        * A tree composed of Consent nodes with edges modeling the dependency
            relationships between them.
        * A `meets_scope_requirements` method is defined to check whether a scope
            requirement, including dependent scope requirements, is satisfied by the
            tree.
    * ``ConsentForest``
        * A collection of all ConsentTrees for a user rooted under a particular client
            (the client that initiated the request for consents).
        * A `meets_scope_requirements` method is defined to check whether a scope
            requirement, including dependent scope requirements, is satisfied by any
            tree in the forest.
"""

from __future__ import annotations

import textwrap
import typing as t
import uuid
from dataclasses import dataclass
from datetime import datetime

from ..parser import ScopeParser
from ..representation import Scope
from ._errors import ConsentParseError, ConsentTreeConstructionError


@dataclass
class Consent:
    """
    Consent Data Object

    This object models:
        * A grant which a user has provided for a client to perform a particular
            scoped operation on their behalf.
        * The consent is conditional on the invocation path leading to the client's
            attempted operation being initiated through a chain of similarly scoped
            operations (consents) defined in the "dependency_path".
    """

    client: uuid.UUID | str
    scope: uuid.UUID | str
    scope_name: str
    id: int
    effective_identity: uuid.UUID | str
    # A list representing the path of consent dependencies leading from a "root consent"
    #   to this. The last element of this list will always be this consent's ID.
    # Downstream dependency relationships may exist but will not be defined here.
    dependency_path: list[int]
    created: datetime
    updated: datetime
    last_used: datetime
    status: str
    allows_refresh: bool
    auto_approved: bool
    atomically_revocable: bool

    @classmethod
    def load(cls, data: t.Mapping[str, t.Any]) -> Consent:
        """
        Load a Consent object from a raw data dictionary.

        :param data: A dictionary containing the raw consent data.
        :raises: ConsentParseError if the data is missing a required key.
        """
        try:
            return cls(
                id=data["id"],
                client=data["client"],
                scope=data["scope"],
                effective_identity=data["effective_identity"],
                dependency_path=data["dependency_path"],
                scope_name=data["scope_name"],
                created=datetime.fromisoformat(data["created"]),
                updated=datetime.fromisoformat(data["updated"]),
                last_used=datetime.fromisoformat(data["last_used"]),
                status=data["status"],
                allows_refresh=data["allows_refresh"],
                auto_approved=data["auto_approved"],
                atomically_revocable=data["atomically_revocable"],
            )
        except KeyError as e:
            raise ConsentParseError(
                f"Failed to load Consent object. Missing required key: {e}.",
                raw_consent=dict(data),
            ) from e

    def __str__(self) -> str:
        client_id = str(self.client)
        return f"Consent [{self.id}]: Client [{client_id}] -> Scope [{self.scope_name}]"


class ConsentForest:
    """
    A ConsentForest is a data structure which models relationships between Consents,
        objects describing explicit access users have granted to particular clients.
    It exists to expose a simple interface for evaluating whether resource server grant
        requirements, as defined by a scope object are satisfied.

    Consents should be retrieved from the AuthClient's ``get_consents`` method.

    Example usage:

    >>> auth_client = AuthClient(...)
    >>> identity_id = ...
    >>> forest = auth_client.get_consents(identity_id).to_forest()
    >>>
    >>> # Check whether the forest meets a scope requirement
    >>> data_access_scope = GCSCollectionScopes(collection_id).data_access
    >>> scope = TransferScopes.all.with_dependency(data_access_scope)
    >>> forest.meets_scope_requirements(scope)


    The following diagram demonstrates a Consent Forest in which a user has consented
        to a client ("CLI") initiating transfers against two collections, both of which
        require a "data_access" dynamic scope.
    Contained Scope String:
        `transfer:all[<collection1>:data_access <collection2>:data_access]`

    .. code-block:: rst

        [Consent A          ]    [Consent B                       ]
        [Client: CLI        ] -> [Client: Transfer                ]
        [Scope: transfer:all]    [Scope: <collection1>:data_access]
                |
                |                [Consent C                       ]
                |--------------> [Client: Transfer                ]
                                 [Scope: <collection2>:data_access]
    """

    def __init__(self, consents: t.Iterable[t.Mapping[str, t.Any] | Consent]) -> None:
        """
        :param consents: An iterable of consent data objects. Typically, this will be
            a ConsentForestResponse retrieved via `auth_client.get_consents(identity)`.
            This iterable may contain either raw consent data as a dict or pre-loaded
            Consents.
        """
        self.nodes = [
            consent if isinstance(consent, Consent) else Consent.load(consent)
            for consent in consents
        ]
        # Build an index on consent id for constant time lookups
        self._node_by_id = {node.id: node for node in self.nodes}

        self.edges = self._compute_edges()
        self.trees = self._build_trees()

    def __str__(self) -> str:
        # indent 4 for inner elements, so that we can put their headings at 2 indent
        trees = textwrap.indent("\n".join(str(t) for t in self.trees), "    ")
        nodes = textwrap.indent("\n".join(str(n) for n in self.nodes), "    ")
        return f"ConsentForest\n  nodes\n{nodes}\n  trees\n{trees}"

    def _compute_edges(self) -> dict[int, set[int]]:
        """
        Compute the edges of the forest mapping parent -> child.

        A consent's parent node id is defined as the penultimate element of the
            consent's dependency path.
        A consent with dependency list of length 1 is a root node (has no parent).
        """
        edges: dict[int, set[int]] = {node.id: set() for node in self.nodes}
        for node in self.nodes:
            if len(node.dependency_path) > 1:
                parent_id = node.dependency_path[-2]
                try:
                    edges[parent_id].add(node.id)
                except KeyError as e:
                    raise ConsentTreeConstructionError(
                        f"Failed to compute forest edges. Missing parent node: {e}.",
                        consents=self.nodes,
                    ) from e
        return edges

    def _build_trees(self) -> list[ConsentTree]:
        """
        Build out the list of trees in the forest.

        A distinct tree is built out for each "root nodes" (nodes with no parents).
        """

        # A node with dependency path length 1 has no parents, so it is a root.
        roots = [node for node in self.nodes if len(node.dependency_path) == 1]
        return [ConsentTree(root.id, self) for root in roots]

    def get_node(self, consent_id: int) -> Consent:
        return self._node_by_id[consent_id]

    def meets_scope_requirements(
        self, scopes: Scope | str | t.Sequence[Scope | str]
    ) -> bool:
        """
        Check whether this consent meets one or more scope requirements.

        A consent forest meets a particular scope requirement if any consent tree
            inside the forest meets the scope requirements.

        :param scopes: A single scope, a list of scopes, or a scope string to check
           against the forest.
        :returns: True if all scope requirements are met, False otherwise.
        """
        for scope in _normalize_scope_types(scopes):
            if not any(tree.meets_scope_requirements(scope) for tree in self.trees):
                return False
        return True


class ConsentTree:
    """
    A tree of Consent nodes with edges modeling the dependency relationships between
        them.

    :raises: ConsentParseError if the tree cannot be constructed due to missing
       consent dependencies.
    """

    def __init__(self, root_id: int, forest: ConsentForest) -> None:
        self.root = forest.get_node(root_id)
        self.nodes = [self.root]
        self._node_by_id = {root_id: self.root}
        self.edges: dict[int, set[int]] = {}

        self._populate_connected_nodes_and_edges(forest)

    def _populate_connected_nodes_and_edges(self, forest: ConsentForest) -> None:
        """
        Populate the nodes and edges of the tree by traversing the forest.

        Nodes/Edges are included in the tree iff they are reachable from the root.
        """
        nodes_to_evaluate = {self.root.id}
        while nodes_to_evaluate:
            consent_id = nodes_to_evaluate.pop()
            consent = forest.get_node(consent_id)
            self.edges[consent.id] = forest.edges[consent.id]

            for child_id in self.edges[consent.id]:
                if child_id not in self._node_by_id:
                    self.nodes.append(forest.get_node(child_id))
                    self._node_by_id[child_id] = forest.get_node(child_id)
                    nodes_to_evaluate.add(child_id)

    def get_node(self, consent_id: int) -> Consent:
        return self._node_by_id[consent_id]

    def meets_scope_requirements(self, scope: Scope) -> bool:
        """
        Check whether this consent tree meets a particular scope requirement.

        :param scope: A Scope requirement to check against the tree.
        """
        return self._meets_scope_requirements_recursive(self.root, scope)

    def _meets_scope_requirements_recursive(self, node: Consent, scope: Scope) -> bool:
        """
        Check recursively whether a consent node meets the scope requirements defined
            by a scope object.
        """
        if node.scope_name != scope.scope_string:
            return False

        for dependent_scope in scope.dependencies:
            for child_id in self.edges[node.id]:
                if self._meets_scope_requirements_recursive(
                    self.get_node(child_id), dependent_scope
                ):
                    # We found a child containing this full dependent scope tree
                    # Move onto the next dependent scope tree
                    break
            else:
                # We didn't find any child containing this full dependent scope tree
                return False
        # We found at least one child containing each full dependent scope tree
        return True

    @property
    def max_depth(self) -> int:
        return self._max_depth_recursive(self.root, 1)

    def _max_depth_recursive(self, node: Consent, depth: int) -> int:
        if len(self.edges[node.id]) == 0:
            return depth
        return max(
            self._max_depth_recursive(self.get_node(child_id), depth + 1)
            for child_id in self.edges[node.id]
        )

    def __str__(self) -> str:
        """Returns a textual representation of the tree to stdout (one line per node)"""
        return self._str_recursive(self.root, 0)

    def _str_recursive(self, node: Consent, tab_depth: int) -> str:
        _str = f"{' ' * tab_depth} - {node}"
        for child_id in self.edges[node.id]:
            _str += "\n" + self._str_recursive(self.get_node(child_id), tab_depth + 2)
        return _str


def _normalize_scope_types(
    scopes: Scope | str | t.Sequence[Scope | str],
) -> list[Scope]:
    """
    Normalize the input scope types into a list of Scope objects.

    Strings are parsed into 1 or more Scopes using `ScopeParser.parse`.

    :param scopes: Some collection of 0 or more scopes as Scope or scope strings.
    :returns: A list of Scope objects.
    """

    if isinstance(scopes, Scope):
        return [scopes]
    elif isinstance(scopes, str):
        return ScopeParser.parse(scopes)
    else:
        scope_list = []
        for scope in scopes:
            if isinstance(scope, str):
                scope_list.extend(ScopeParser.parse(scope))
            else:
                scope_list.append(scope)
        return scope_list



from __future__ import annotations

import typing as t

import requests

from globus_sdk import exc


class SearchAPIError(exc.GlobusAPIError):
    """
    Error class for the Search API client. In addition to the
    inherited instance variables, provides ``error_data``.

    :ivar error_data: Additional object returned in the error response. May be
                      a dict, list, or None.
    """

    def __init__(self, r: requests.Response) -> None:
        self.error_data: dict[str, t.Any] | None = None
        super().__init__(r)

    def _post_parse_hook(self) -> bool:
        self.error_data = self._dict_data.get("error_data")
        return True



import pytest
import requests

import globus_sdk
from globus_sdk.testing import RegisteredResponse
from globus_sdk.transport import RequestCallerInfo, RetryConfig


@pytest.mark.parametrize("error_status", [500, 429, 502, 503, 504])
def test_retry_on_transient_error(client, mocksleep, error_status):
    RegisteredResponse(
        path="https://foo.api.globus.org/bar", status=error_status, body="Uh-oh!"
    ).add()
    RegisteredResponse(path="https://foo.api.globus.org/bar", json={"baz": 1}).add()

    # no sign of an error in the client
    res = client.get("/bar")
    assert res.http_status == 200
    assert res["baz"] == 1

    # there was a sleep (retry was triggered)
    mocksleep.assert_called_once()


def test_retry_disabled_via_tune(client, mocksleep):
    RegisteredResponse(
        path="https://foo.api.globus.org/bar", status=500, body="Uh-oh!"
    ).add()
    RegisteredResponse(path="https://foo.api.globus.org/bar", json={"baz": 1}).add()

    # the error is seen by the client (automatic retry does not hide it)
    with pytest.raises(globus_sdk.GlobusAPIError) as excinfo:
        with client.retry_config.tune(max_retries=0):
            client.get("/bar")
    assert excinfo.value.http_status == 500

    # there was no sleep (retry was not triggered)
    mocksleep.assert_not_called()


def test_retry_disabled_via_init_param(client_class, mocksleep):
    RegisteredResponse(
        path="https://foo.api.globus.org/bar", status=500, body="Uh-oh!"
    ).add()
    RegisteredResponse(path="https://foo.api.globus.org/bar", json={"baz": 1}).add()
    client = client_class(retry_config=RetryConfig(max_retries=0))

    # the error is seen by the client (automatic retry does not hide it)
    with pytest.raises(globus_sdk.GlobusAPIError) as excinfo:
        client.get("/bar")
    assert excinfo.value.http_status == 500

    # there was no sleep (retry was not triggered)
    mocksleep.assert_not_called()


def test_retry_disabled_via_init_param_but_enabled_via_tune(client_class, mocksleep):
    RegisteredResponse(
        path="https://foo.api.globus.org/bar", status=500, body="Uh-oh!"
    ).add()
    RegisteredResponse(path="https://foo.api.globus.org/bar", json={"baz": 1}).add()
    client = client_class(retry_config=RetryConfig(max_retries=0))

    # no sign of an error in the client if we "turn it back on"
    with client.retry_config.tune(max_retries=1):
        res = client.get("/bar")
    assert res.http_status == 200
    assert res["baz"] == 1

    # there was a sleep (retry was triggered)
    mocksleep.assert_called_once()


def test_retry_on_network_error(client, mocksleep):
    # set the response to be a requests NetworkError -- responses will raise the
    # exception when the call is made
    RegisteredResponse(
        path="https://foo.api.globus.org/bar",
        body=requests.ConnectionError("foo-err"),
    ).add()
    RegisteredResponse(path="https://foo.api.globus.org/bar", json={"baz": 1}).add()

    # no sign of an error in the client
    res = client.get("/bar")
    assert res.http_status == 200
    assert res["baz"] == 1

    # there was a sleep (retry was triggered)
    mocksleep.assert_called_once()


@pytest.mark.parametrize("num_errors,expect_err", [(5, False), (6, True), (7, True)])
def test_retry_limit(client, mocksleep, num_errors, expect_err):
    # N errors followed by a success
    for _i in range(num_errors):
        RegisteredResponse(
            path="https://foo.api.globus.org/bar", status=500, body="Uh-oh!"
        ).add()
    RegisteredResponse(path="https://foo.api.globus.org/bar", json={"baz": 1}).add()

    if expect_err:
        with pytest.raises(globus_sdk.GlobusAPIError):
            client.get("/bar")
    else:
        # no sign of an error in the client
        res = client.get("/bar")
        assert res.http_status == 200
        assert res["baz"] == 1

    # default num retries = 5
    assert mocksleep.call_count == min(num_errors, 5)


def test_transport_retry_limit(client, mocksleep):
    # this limit is a safety to protect against a bad policy causing infinite retries
    client.retry_config.max_retries = 2

    for _i in range(3):
        RegisteredResponse(
            path="https://foo.api.globus.org/bar", status=500, body="Uh-oh!"
        ).add()
    RegisteredResponse(path="https://foo.api.globus.org/bar", json={"baz": 1}).add()

    with pytest.raises(globus_sdk.GlobusAPIError):
        client.get("/bar")

    assert mocksleep.call_count == 2


def test_bad_max_retries_causes_error(client):
    # this test exploits the fact that we loop to (max_retries + 1) in order
    # to ensure that no requests are ever sent
    # the transport should throw an error in this case, since it doesn't have a response
    # value to return
    client.retry_config.max_retries = -1

    with pytest.raises(ValueError):
        client.get("/bar")


def test_persistent_connection_error(client):
    for _i in range(6):
        RegisteredResponse(
            path="https://foo.api.globus.org/bar",
            body=requests.ConnectionError("foo-err"),
        ).add()
    RegisteredResponse(path="https://foo.api.globus.org/bar", json={"baz": 1}).add()

    with pytest.raises(globus_sdk.GlobusConnectionError):
        client.get("/bar")


def test_no_retry_401_no_authorizer(client):
    RegisteredResponse(
        path="https://foo.api.globus.org/bar", status=401, body="Unauthorized"
    ).add()
    RegisteredResponse(path="https://foo.api.globus.org/bar", json={"baz": 1}).add()

    # error gets raised in client (no retry)
    with pytest.raises(globus_sdk.GlobusAPIError) as excinfo:
        client.get("/bar")
    assert excinfo.value.http_status == 401


def test_retry_with_authorizer(client):
    RegisteredResponse(
        path="https://foo.api.globus.org/bar", status=401, body="Unauthorized"
    ).add()
    RegisteredResponse(path="https://foo.api.globus.org/bar", json={"baz": 1}).add()

    # an authorizer class which does nothing but claims to support handling of
    # unauthorized errors
    dummy_authz_calls = []

    class DummyAuthorizer(globus_sdk.authorizers.GlobusAuthorizer):
        def get_authorization_header(self):
            dummy_authz_calls.append("set_authz")
            return "foo"

        def handle_missing_authorization(self):
            dummy_authz_calls.append("handle_missing")
            return True

    authorizer = DummyAuthorizer()
    client.authorizer = authorizer

    # no sign of an error in the client
    res = client.get("/bar")
    assert res.http_status == 200
    assert res["baz"] == 1

    # ensure that setting authz was called twice (once for each request)
    # and that between the two calls, handle_missing_authorization was called once
    assert dummy_authz_calls == ["set_authz", "handle_missing", "set_authz"]


def test_no_retry_with_authorizer_no_handler(client):
    RegisteredResponse(
        path="https://foo.api.globus.org/bar", status=401, body="Unauthorized"
    ).add()
    RegisteredResponse(path="https://foo.api.globus.org/bar", json={"baz": 1}).add()

    # an authorizer class which does nothing and does not claim to handle
    # unauthorized errors
    dummy_authz_calls = []

    class DummyAuthorizer(globus_sdk.authorizers.GlobusAuthorizer):
        def get_authorization_header(self):
            dummy_authz_calls.append("set_authz")
            return "foo"

        def handle_missing_authorization(self):
            dummy_authz_calls.append("handle_missing")
            return False

    authorizer = DummyAuthorizer()
    client.authorizer = authorizer

    # error gets raised in client (no retry)
    with pytest.raises(globus_sdk.GlobusAPIError) as excinfo:
        client.get("/bar")
    assert excinfo.value.http_status == 401

    # only two calls, single setting of authz and a call to ask about handling the error
    assert dummy_authz_calls == ["set_authz", "handle_missing"]


def test_retry_with_authorizer_persistent_401(client):
    RegisteredResponse(
        path="https://foo.api.globus.org/bar", status=401, body="Unauthorized"
    ).add()
    RegisteredResponse(
        path="https://foo.api.globus.org/bar", status=401, body="Unauthorized"
    ).add()
    RegisteredResponse(path="https://foo.api.globus.org/bar", json={"baz": 1}).add()

    # an authorizer class which does nothing but claims to support handling of
    # unauthorized errors
    dummy_authz_calls = []

    class DummyAuthorizer(globus_sdk.authorizers.GlobusAuthorizer):
        def get_authorization_header(self):
            dummy_authz_calls.append("set_authz")
            return "foo"

        def handle_missing_authorization(self):
            dummy_authz_calls.append("handle_missing")
            return True

    authorizer = DummyAuthorizer()
    client.authorizer = authorizer

    # the error gets raised in this case because it persists -- the authorizer only gets
    # one chance to resolve the issue
    with pytest.raises(globus_sdk.GlobusAPIError) as excinfo:
        client.get("/bar")
    assert excinfo.value.http_status == 401

    # ensure that setting authz was called twice (once for each request)
    # and that between the two calls, handle_missing_authorization was called once
    # but the handler should not be called a second time because the 401 repeated
    assert dummy_authz_calls == ["set_authz", "handle_missing", "set_authz"]


def test_transport_caller_info_with_retry(client):
    RegisteredResponse(
        path="https://foo.api.globus.org/bar", status=401, body="Unauthorized"
    ).add()
    RegisteredResponse(path="https://foo.api.globus.org/bar", json={"baz": 1}).add()

    dummy_authz_calls = []

    class DummyAuthorizer(globus_sdk.authorizers.GlobusAuthorizer):
        def get_authorization_header(self):
            dummy_authz_calls.append("set_authz")
            return "foo"

        def handle_missing_authorization(self):
            dummy_authz_calls.append("handle_missing")
            return True

    authorizer = DummyAuthorizer()
    caller_info = RequestCallerInfo(
        retry_config=client.retry_config, authorizer=authorizer
    )

    # Test direct transport usage with caller_info
    response = client.transport.request(
        "GET", "https://foo.api.globus.org/bar", caller_info=caller_info
    )

    assert response.status_code == 200
    # Verify that the authorizer was used for both authorization and retry handling
    assert dummy_authz_calls == ["set_authz", "handle_missing", "set_authz"]



import json
import uuid

import pytest

from globus_sdk import BatchMembershipActions, GroupRole
from globus_sdk.testing import RegisteredResponse, get_last_request, load_response
from tests.common import register_api_route_fixture_file


def test_approve_pending(groups_manager):
    register_api_route_fixture_file(
        "groups",
        "/v2/groups/d3974728-6458-11e4-b72d-123139141556",
        "approve_pending.json",
        method="POST",
    )

    res = groups_manager.approve_pending(
        "d3974728-6458-11e4-b72d-123139141556", "ae332d86-d274-11e5-b885-b31714a110e9"
    )
    assert res.http_status == 200

    data = res.data
    assert isinstance(data, dict)
    assert "approve" in data
    assert data["approve"][0]["status"] == "active"


@pytest.mark.parametrize("role", (GroupRole.admin, GroupRole.member, "member", "admin"))
def test_add_member(groups_manager, role):
    rolestr = role if isinstance(role, str) else role.value
    register_api_route_fixture_file(
        "groups",
        "/v2/groups/d3974728-6458-11e4-b72d-123139141556",
        "add_member.json",
        method="POST",
    )

    res = groups_manager.add_member(
        "d3974728-6458-11e4-b72d-123139141556",
        "ae332d86-d274-11e5-b885-b31714a110e9",
        role=role,
    )
    assert res.http_status == 200

    data = res.data
    assert isinstance(data, dict)
    assert "add" in data
    assert data["add"][0]["status"] == "active"
    # FIXME: this should be the line
    #   assert data["add"][0]["role"] == rolestr
    # but the response is fixed right now
    assert data["add"][0]["role"] == "admin"

    req = get_last_request()
    req_body = json.loads(req.body)
    assert req_body["add"][0]["role"] == rolestr


@pytest.mark.parametrize("role", (GroupRole.admin, GroupRole.member, "member", "admin"))
def test_batch_action_payload(groups_client, role):
    group_id = str(uuid.uuid1())
    load_response(
        RegisteredResponse(
            service="groups", method="POST", path=f"/v2/groups/{group_id}", json={}
        )
    )
    rolestr = role if isinstance(role, str) else role.value

    batch_action = (
        BatchMembershipActions()
        .accept_invites(uuid.uuid1())
        .add_members(
            [uuid.uuid1(), uuid.uuid1()],
            role=role,
        )
        .change_roles("admin", [uuid.uuid1(), uuid.uuid1()])
        .invite_members([uuid.uuid1(), uuid.uuid1()])
        .join([uuid.uuid1(), uuid.uuid1()])
    )

    assert "add" in batch_action
    assert len(batch_action["add"]) == 2
    assert all(member["role"] == role for member in batch_action["add"])

    assert "accept" in batch_action
    assert len(batch_action["accept"]) == 1

    assert "change_role" in batch_action
    assert len(batch_action["change_role"]) == 2
    for change_role in batch_action["change_role"]:
        assert change_role["role"] == "admin"

    assert "invite" in batch_action
    assert len(batch_action["invite"]) == 2

    assert "join" in batch_action
    assert len(batch_action["invite"]) == 2

    # send the request and confirm that the data is serialized correctly
    groups_client.batch_membership_action(group_id, batch_action)
    req = get_last_request()
    req_body = json.loads(req.body)
    # role should be stringified if it was an enum member
    assert all(member["role"] == rolestr for member in req_body["add"])
    # UUIDs should have been stringified
    for action in ["add", "accept", "invite", "join"]:
        assert all(isinstance(value["identity_id"], str) for value in req_body[action])



import pytest

from globus_sdk import __version__
from globus_sdk.transport import GlobusClientInfo


def make_empty_clientinfo():
    # create a clientinfo with no contents, as a starting point for tests
    obj = GlobusClientInfo()
    obj.clear()
    return obj


def parse_clientinfo(header):
    """
    Sample parser.

    Including this in the testsuite not only validates the mechanical implementation of
    X-Globus-Client-Info, but also acts as a safety check that we've thought through the
    ability of consumers to parse this data.
    """
    mappings = {}
    for segment in header.split(";"):
        segment_dict = {}

        segment = segment.strip()
        elements = segment.split(",")

        for element in elements:
            if "=" not in element:
                raise ValueError(
                    f"Bad X-Globus-Client-Info element: '{element}' in '{header}'"
                )
            key, _, value = element.partition("=")
            if "=" in value:
                raise ValueError(
                    f"Bad X-Globus-Client-Info element: '{element}' in '{header}'"
                )
            if key in segment_dict:
                raise ValueError(
                    f"Bad X-Globus-Client-Info element: '{element}' in '{header}'"
                )
            segment_dict[key] = value
        if "product" not in segment_dict:
            raise ValueError(
                "Bad X-Globus-Client-Info segment missing product: "
                f"'{segment}' in '{header}'"
            )
        product = segment_dict["product"]
        if product in mappings:
            raise ValueError(
                "Bad X-Globus-Client-Info header repeats product: "
                f"'{product}' in '{header}'"
            )
        mappings[product] = segment_dict
    return mappings


def test_clientinfo_bool():
    # base clientinfo starts with the SDK version and should bool true
    info = GlobusClientInfo()
    assert bool(info) is True
    # but we can clear it and it will bool False
    info.infos = []
    assert bool(info) is False


@pytest.mark.parametrize(
    "value, expect_str",
    (
        ("x=y", "x=y"),
        ("x=y,omicron=iota", "x=y,omicron=iota"),
        ({"x": "y"}, "x=y"),
        ({"x": "y", "alpha": "b01"}, "x=y,alpha=b01"),
    ),
)
def test_format_of_simple_item(value, expect_str):
    info = make_empty_clientinfo()
    info.add(value)
    assert info.format() == expect_str


@pytest.mark.parametrize(
    "values, expect_str",
    (
        (("x=y",), "x=y"),
        (("x=y", "alpha=b01,omicron=iota"), "x=y;alpha=b01,omicron=iota"),
    ),
)
def test_format_of_multiple_items(values, expect_str):
    info = make_empty_clientinfo()
    for value in values:
        info.add(value)
    assert info.format() == expect_str


def test_clientinfo_parses_as_expected():
    info = GlobusClientInfo()
    info.add("alpha=b01,product=my-cool-tool")
    header_str = info.format()

    parsed = parse_clientinfo(header_str)
    assert parsed == {
        "python-sdk": {
            "product": "python-sdk",
            "version": __version__,
        },
        "my-cool-tool": {
            "product": "my-cool-tool",
            "alpha": "b01",
        },
    }


def test_client_info_can_write_back_via_callback():
    myvalue = ""

    def onupdate(info):
        nonlocal myvalue
        myvalue = info.format()

    info = GlobusClientInfo(update_callback=onupdate)

    # initializing with the callback does not make it fire
    # the value is unchanged
    assert myvalue == ""

    segment = "version=1.0.1,product=my-cool-tool"
    # now, add something and make sure it rendered back into the value
    # (along with python-sdk info)
    info.add(segment)

    # our new segment is visible
    assert segment in myvalue
    # but other values (the default, python-sdk version!) are also there
    assert myvalue != segment



from __future__ import annotations

import logging
import typing as t
import uuid

from globus_sdk._missing import MISSING, MissingType
from globus_sdk._payload import GlobusPayload

log = logging.getLogger(__name__)


class BookmarkCreateDocument(GlobusPayload):
    """
    Convenience class for constructing a bookmark document to use as the
    ``data`` parameter to
    :meth:`~globus_sdk.TransferClientV2.create_bookmark`.
    """

    def __init__(
        self,
        collection: uuid.UUID | str,
        name: str,
        path: str,
        *,
        pinned: bool | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        log.debug("Creating a new BookmarkCreateDocument object")

        relationships = {
            "collection": {
                "data": {
                    "type": "Collection",
                    "id": collection,
                }
            }
        }

        attributes = {
            "name": name,
            "path": path,
            "pinned": pinned,
        }

        if additional_fields is not None:
            attributes.update(additional_fields)

        self["data"] = {
            "type": "Bookmark",
            "attributes": attributes,
            "relationships": relationships,
        }


class BookmarkUpdateDocument(GlobusPayload):
    """
    Convenience class for constructing a bookmark document to use as the
    ``data`` parameter to
    :meth:`~globus_sdk.TransferClientV2.update_bookmark`.
    """

    def __init__(
        self,
        name: str,
        path: str,
        *,
        pinned: bool | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        log.debug("Creating a new BookmarkUpdateDocument")

        attributes = {
            "name": name,
            "path": path,
            "pinned": pinned,
        }

        if additional_fields is not None:
            attributes.update(additional_fields)

        self["data"] = {
            "type": "Bookmark",
            "attributes": attributes,
        }



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

CREDENTIAL_ID = "af43d884-64a1-4414-897a-680c32374439"

RESPONSES = ResponseSet(
    metadata={"id": CREDENTIAL_ID},
    default=RegisteredResponse(
        service="gcs",
        path=f"/user_credentials/{CREDENTIAL_ID}",
        method="DELETE",
        json={
            "DATA_TYPE": "result#1.0.0",
            "code": "success",
            "data": [],
            "detail": "success",
            "has_next_page": False,
            "http_response_code": 200,
            "message": f"Deleted User Credential {CREDENTIAL_ID}",
        },
    ),
)
